/*
	TinyCAD program for schematic capture
	Copyright 1994/1995/2002 Matt Pyne.

	This program is free software; you can redistribute it and/or
	modify it under the terms of the GNU Lesser General Public
	License as published by the Free Software Foundation; either
	version 2.1 of the License, or (at your option) any later version.

	This program is distributed in the hope that it will be useful,
	but WITHOUT ANY WARRANTY; without even the implied warranty of
	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
	Lesser General Public License for more details.

	You should have received a copy of the GNU Lesser General Public
	License along with this library; if not, write to the Free Software
	Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
*/



#ifndef __CONTEXT_H__
#define __CONTEXT_H__

#include "option.h"
#include "ArcPoint.h"

#include <map>

#define STRLEN		256

#define MAX_FONTS	8
#define MAX_PENS	20
#define MAX_BRUSHES	4


// The class that performs the snap to grid, when necessary
class TransformSnap
{
protected:
	LONG	grid;				// The spacing on the grid
	BOOL	GridSnap;			// Should we snap to grid?
public:
	// Set the current grid settings
	void SetAccurateGrid(int NewGrid) { grid = NewGrid; }
	void SetGridSnap(BOOL NewSnap) { GridSnap = NewSnap; }

	// Get the current settings
	int GetAccurateGrid() { return grid; }	
	int GetGrid() { return grid / 10; }	
	BOOL GetGridSnap() { return GridSnap; }

	// Snap co-ords to the grid
	int Snap(LONG a);
	CPoint Snap(CPoint p) { return CPoint(Snap(p.x),Snap(p.y)); }

	// The constructor
	TransformSnap();

	void SaveXML( CXMLWriter &xml )
	{
		xml.addTag( "GRID" );
		xml.addAttribute( "spacing", grid );
		xml.addAttribute( "snap", GridSnap );
		xml.closeTag();
	}

	void LoadXML( CXMLReader &xml )
	{
		xml.getAttribute( "spacing", grid );
		xml.getAttribute( "snap", GridSnap );
	}


	void Load( CStream &theArchive, BOOL keep )
	{
		if (keep)
		{
			theArchive >> grid >> GridSnap;
		}
		else
		{
			LONG dummy1;
			BOOL dummy2;
			theArchive >> dummy1 >> dummy2;
		}
	}

};

// The Transform context, which combines transform and scalings
class Transform {

	LONG	zoom;				// The zoom of the drawing
	double	scaling_x;			// The scaling of a method part
	double	scaling_y;			// The scaling of a method part
	int		x,y;				// The coord of the top left hand corner (The origin)
	int		rotmir;				// The rotation and mirroring


	BOOL	IsYUp;				// Should we invert the y-axis?
	int		BoundY;				// The Y bound - used when y is up

	// Internal scaling functions
	int ScaleX(int a) { return intScale(a * scaling_x -x); }
	int ScaleY(int a) { return IsYUp ? -BoundY-intScale(a * scaling_y -y) : intScale(a * scaling_y -y); }

	double ScaleX(double a) { return doubleScale(a * scaling_x -x); }
	double ScaleY(double a) { return IsYUp ? -BoundY-doubleScale(a * scaling_y -y) : doubleScale(a * scaling_y -y); }

public:

	// Change the settings
	void SetZoom(LONG NewZoom) { zoom = (NewZoom << 8)/100; }
	void SetOriginX(int NewX) { x = NewX; }
	void SetOriginY(int NewY) { y = NewY; }
	void SetOrigin(CPoint p) { x = p.x; y = p.y; }
	void SetYisup(int newBoundY) { IsYUp = TRUE; BoundY = newBoundY; }
	void SetScaling( int rot, double x, double y );


	
	// Change the transform by rotating it 
	CPoint SetTRM(CPoint,CPoint,int);
	void EndTRM(CPoint);
	int RotateDir(int);

	CPoint GetOrigin() { return CPoint(x,y); }
	LONG GetZoom() { return (zoom*100)>>8; }
	void GetScaling( int rot, double &x, double &y );
	BOOL GetIsYUp() { return IsYUp; }

	int GetRotMir() { return rotmir; } 
	

	// Convert from internal coord into window coords
	CPoint Scale(CPoint);
	CPoint Scale(CDPoint);
	double doubleScale(double a);
	int intScale(LONG a);
	int Scale(int a) { return intScale(a)>0 ? intScale(a) : 1; }
	CRect Scale(CRect r) { CPoint a = Scale(CPoint(r.left,r.top)); CPoint b = Scale(CPoint(r.right,r.bottom)); return CRect(a.x,a.y,b.x,b.y); }

	// Convert from window coords into internal coords
	CPoint DeScale(TransformSnap &s,CPoint p) { return CPoint(DeScaleX(s,p.x),DeScaleY(s,p.y)); }
	int intDeScale(LONG a);
	int DeScaleX(TransformSnap &s, LONG a) { return s.Snap(DeScaleX(a)); }
	int DeScaleY(TransformSnap &s, LONG a) { return s.Snap(DeScaleY(a)); }

	CPoint DeScale(CPoint p) { return CPoint(DeScaleX(p.x),DeScaleY(p.y)); }
	int DeScaleX(LONG a) { return (((a<<8)/zoom)+x); }
	int DeScaleY(LONG a) { return (((a<<8)/zoom)+y); }

	// The constructor
	Transform();
};

class CTinyCadDoc;

inline bool operator<(const LOGFONT&a,const LOGFONT&b)
{

	if (a.lfEscapement < b.lfEscapement)
		return true;
	if (a.lfEscapement == b.lfEscapement)
	{
		if (a.lfHeight < b.lfHeight )
			return true;

		if (a.lfHeight == b.lfHeight)
		{
			if (a.lfWidth < b.lfWidth )
				return true;
		}
	}

	return false;
}



// The Context class, a sort of extented DC
class Context {

	BOOL		deleteDC;			// Delete the DC when destructor called
	BOOL		allBlack;			// Make all colours Black
	BOOL		allGrey;			// Make all colours lighter

	CSize		GetTextExtentI(CString);


	// The selectable objects into the DC

	// struct to hold font information
	typedef std::map<LOGFONT,CFont*> font_map;
	font_map	m_fonts;
	CFont		*theDefaultFont;

	int			nextNewPen;
	int			thePenWidth[MAX_PENS], thePenStyle[MAX_PENS];
	LONG		thePenColour[MAX_PENS];
	CPen		*thePen[MAX_PENS], *theDefaultPen;

	int			nextNewBrush;
	LONG		theBrushColour[MAX_BRUSHES];
	int			theBrushIndex[MAX_BRUSHES];
	CBrush		*theBrush[MAX_BRUSHES], *theDefaultBrush;

public:
	CDC			*dc;				// The device context associated with this Context
	Transform	theTransform;		// The transform parameter associated with this Context
	CWnd		*theWindow;			// The window associated with this device


	Context(CWnd *,Transform);		// Construction from window
	Context(CDC *,Transform, CWnd *pWnd = NULL);		// Construction from CDC

	virtual ~Context();							// The destructor


	// Select objects to paint with
	BOOL SelectPen(int Style,int Width,LONG Colour, paint_options options = draw_normal);
	BOOL SelectPen(LineStyle *pStyle, paint_options options );
	BOOL SelectBrush(LONG Colour, int index = 0);
	BOOL SelectBrush(FillStyle *pStyle) { return SelectBrush( pStyle->Colour, pStyle->Index); }
	BOOL SelectBrush() { dc->SelectStockObject(HOLLOW_BRUSH); return TRUE; }
	BOOL SelectFont(LOGFONT &lf, int Rotation, BOOL Scale = TRUE );


	// Change the parameters
	void SetBlack(BOOL NewBlack) { allBlack = NewBlack; }
	void SetROP2(int r) { dc->SetROP2(r); }
	void SetTextColor(LONG Colour) { dc->SetTextColor(allBlack ? RGB(0,0,0) : Colour); }
	void SetTextAlign(int a) { dc->SetTextAlign(a); }
	void SetBkMode(int a) { dc->SetBkMode(a); }
	void SetGreyMode(BOOL a) { allGrey = a; }

	// Get the parameters
	BOOL GetBlack() { return allBlack; }
	CRect GetUpdateRegion();
	BOOL IsMetaFile() { return theTransform.GetIsYUp(); }
	

	// Misc operations
	void Invalidate() 
	{ 
		if (theWindow!=NULL) theWindow->Invalidate(); 
	}

	void InvalidateRect( CRect r, BOOL erase ,int grow )
	{
		if (theWindow != NULL)
		{
			if (r.top > r.bottom)
			{
				int q = r.top;
				r.top = r.bottom;
				r.bottom = q;
			}
			if (r.left > r.right)
			{
				int q = r.right;
				r.right = r.left;
				r.left = q;
			}

			r.left -= grow;
			r.top -= grow;
			r.bottom += grow;
			r.right += grow;

			theWindow->InvalidateRect( theTransform.Scale(r), erase );
		}
	}

  
	// The drawing primitives
	void Ellipse(CRect r) { dc->Ellipse(theTransform.Scale(r)); }
	void Rectangle(CRect r) { dc->Rectangle(theTransform.Scale(r)); }
	void Polyline(pointCollection &points, CPoint offset, FillStyle *pStyle);
	void LineTo(CPoint p) { dc->LineTo(theTransform.Scale(p)); }
	void MoveTo(CPoint p) { dc->MoveTo(theTransform.Scale(p)); }
	void SetPixel(CPoint p,int c) { p = theTransform.Scale(p); dc->SetPixel(p.x,p.y,c); }
	void Arc( LPCRECT lpRect, POINT ptStart, POINT ptEnd );
	void QuaterArc(CPoint, CPoint);
	void PolyBezier(POINT *pts,int);
	void TextOut(CString,CPoint,paint_options,int=3);
	void TextOut(int x,int y,const char *t) { CPoint r = theTransform.Scale(CPoint(x,y)); dc->TextOut(r.x,r.y,t); }
	void TextOut(int x,int y, int width, const char *t);
	CSize GetTextExtent(CString s) { CSize r = GetTextExtentI(s); return CSize(theTransform.intDeScale(r.cx),theTransform.intDeScale(r.cy)); };
	void SelectDefault();
};


#endif
