/*
	TinyCAD program for schematic capture
	Copyright 1994/1995/2002,2003 Matt Pyne.

	This program is free software; you can redistribute it and/or
	modify it under the terms of the GNU Lesser General Public
	License as published by the Free Software Foundation; either
	version 2.1 of the License, or (at your option) any later version.

	This program is distributed in the hope that it will be useful,
	but WITHOUT ANY WARRANTY; without even the implied warranty of
	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
	Lesser General Public License for more details.

	You should have received a copy of the GNU Lesser General Public
	License along with this library; if not, write to the Free Software
	Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
*/

#include "stdafx.h"
#include "TinyCadView.h"
#include "diag.h"
#include "colour.h"
#include "layer.h"
#include "option.h"
#include "special.h"
#include "revision.h"
#include "TinyCad.h"

extern Option theOptions;

// Write the header to a file
void HeaderStamp(CStream &theArchive)
{
	CString name = NAME;
  theArchive << name  << (BYTE)FILE_REVISION << (WORD)FILE_MAGIC;
}


void CTinyCadDoc::SaveXML(CXMLWriter& xml, BOOL Details , BOOL SaveSelect)
{
	SaveXML( xml, m_drawing, Details, SaveSelect );
}

void CTinyCadDoc::SaveXML(CXMLWriter &xml, drawingCollection &drawing, BOOL Details, BOOL SaveSelect, BOOL SaveResources )
{
  // Write the objects to the file
  try
  {
  	if (Details) 
	{
		CString comment;

		comment.Format( "This file was written by TinyCAD %s\n"
						"If you wish to view this file go to http://tinycad.sourceforge.net to\n"
						"download a copy.",
						static_cast<CTinyCadApp*>(AfxGetApp())->get_version() );

		xml.addComment( comment );
	}
	  xml.addTag("TinyCAD");

	// If necessary write the header stamp
	if (Details) 
	{
		xml.addTag("DETAILS");
		int Shows = 0;
		if (ShowDetails)
		{
			Shows |= 1;
		}
		if (ShowRulers)
		{
			Shows |= 2;
		}

		xml.addTag("Size");
		xml.addAttribute( "width", MaxXSize );
		xml.addAttribute( "height", MaxYSize );
		xml.closeTag();

		xml.addTag("TITLE", Title );
		xml.addTag("AUTHOR", Author );
		xml.addTag("REVISION", Revision );
		xml.addTag("DOCNUMBER", DocNumber );
		xml.addTag("ORGANISATION", Organisation );
		xml.addTag("SHEETS", Sheets );
		xml.addTag("SHOWS", Shows ); 
		xml.addTag("DATE", DateStr );
		m_snap.SaveXML( xml );

		xml.closeTag();
	}


	if (SaveResources)
	{
		// Find out which resources are in use
		drawingIterator it = drawing.begin();
		while (it != drawing.end()) 
		{
			(*it)->TagResources();
			++ it;
		}

		// Save the resource details details
		theOptions.SaveFontsXML(xml);
		theOptions.SaveStylesXML(xml);
		theOptions.SaveFillStylesXML(xml);
		theOptions.SaveMetaFilesXML(xml);
	}

	// Only save the symbols if we are not saving
	// to a library or the header...
	if (Details)
	{
		theOptions.SaveSymbolsXML(xml);
	}

	if (Details)
	{
		theOptions.SaveOptionsXML(xml);
	}

	  
	drawingIterator it = drawing.begin();
	while (it != drawing.end()) 
	{
		CDrawingObject *obj = *it;

		if (obj->GetType() != xError
			&& 	(Details || !obj->IsConstruction()) 
			&&  (!SaveSelect || IsSelected( obj )) )
		{

			// Now save the actual object data
			obj->SaveXML(xml);
		}

		++ it;
	}

	xml.closeTag();

  }
  catch ( CException *e) 
  {
	// Could not save the file properly
    e->ReportError();
    e->Delete();
  }
}




// Read back the header stamp and check it is OK to read this file
BOOL CheckHeaderStamp(CStream &theArchive, BOOL ReportError)
{
  BYTE tp=-1,revision;
  WORD magic;
  CString name;

  theArchive >> name >> revision >> magic;
  if (name!=NAME || magic!=FILE_MAGIC) {
	// Not written by this program
	  if (ReportError)
	  {
		Message(IDS_ABORTNOTMINE,MB_ICONEXCLAMATION);
	  }
	return FALSE;
  }

  if (revision>FILE_REVISION) {
	// Written by a more advanced version of this program
	  if (ReportError)
	  {
			Message(IDS_ABORTVERSION,MB_ICONEXCLAMATION);
	  }
	return FALSE;
  }

  return TRUE;
}



////// Load a design from a file, loaded design will be selected //////
BOOL CTinyCadDoc::ReadFile(CStream &theArchive, BOOL Details, drawingCollection &drawing)
{
  try
  {
	// Save the old layer setting
	CDrawingObject *obj = NULL;
	BYTE tp=-1;

	LONG pos = theArchive.GetPos();

  	if (!CheckHeaderStamp(theArchive, FALSE))
	{
		// Perhaps this is XML?
		theArchive.Seek(pos);
		CXMLReader xml( &theArchive );
		return ReadFileXML( xml, Details, drawing, FALSE );
	}

	while (tp!=xEndFile) {
		theArchive >> tp;
		switch (tp) {
			case xDesignInformation:
				// Only read this information if necessary
				if (Details) {
					CTime Date;
					LONG MaxXLong, MaxYLong;
					BYTE dummyb;
				  	theArchive >> MaxXLong >> MaxYLong
						>> Title >> Author >> Revision >> DocNumber >> Organisation >> Sheets >> dummyb >> Date;
					MaxXSize = MaxXLong;
					MaxYSize = MaxYLong;
					ShowDetails = dummyb;
					DateStr = Date.Format("%d %B %Y");

				} else {
					CString dummy;
					LONG dummyi;
					BYTE dummyb;
					CTime dummyt;
					theArchive >> dummyi >> dummyi
						>> dummy >> dummy >> dummy >> dummy >> dummy >> dummy >> dummyb >> dummyt;
				}
				obj = NULL;
				break;
			case xDesignInformation2:
				if (Details) {
					LONG MaxXLong, MaxYLong;
					BYTE Shows;
				  	theArchive >> MaxXLong >> MaxYLong
						>> Title >> Author >> Revision >> DocNumber >> Organisation >> Sheets >> Shows >> DateStr;
					MaxXSize = MaxXLong;
					MaxYSize = MaxYLong;
					ShowRulers = Shows & 2;
					ShowDetails = Shows & 1;
				} else {
					CString dummy;
					LONG dummyi;
					BYTE dummyb;
					theArchive >> dummyi >> dummyi
						>> dummy >> dummy >> dummy     >> dummy     >> dummy        >> dummy  >> dummyb >> dummy;
				}
				m_snap.Load( theArchive, Details );
				break;
			case xOptions:
				theOptions.LoadOptions( theArchive, Details );
				obj = NULL;
				break;
			case xFont:
				theOptions.LoadFonts(theArchive);
				obj = NULL;
				break;
			case xLineStyle:
				theOptions.LoadStyles(theArchive);
				obj = NULL;
				break;
			case xFillStyle:
				theOptions.LoadFillStyles(theArchive);
				obj = NULL;
				break;
			case xMetaFiles:
				theOptions.LoadMetaFiles(theArchive);
				obj = NULL;
				break;
			case xSymbols:
				theOptions.LoadSymbols(theArchive);
				obj = NULL;
				break;
			case xLayerTable:
				{
					// Read in the number of Layers in this list
					CString theNewName;
					WORD NumberOfResources;
					theArchive >> NumberOfResources;

					hRESOURCE OldResourceNumber = 0;

					while (NumberOfResources > 0) {
						theArchive >> OldResourceNumber;
						theArchive >> theNewName;
						NumberOfResources--;
					}
					obj = NULL;
				}
				break;
			case xRuler:
				obj = new CDrawRuler(this,FALSE);
				break;
			case xBus:
				obj = new CDrawLine(this,xBus);
				break;
			case xWire:
				obj = new CDrawLine(this,xWire);
				break;
			case xBusName:
				obj = new CDrawText(this,xBusName);
				break;
			case xBusNameEx:
				obj = new CDrawText(this,xBusNameEx);
				break;
			case xLabel:  	// The Old Label type
				obj = new CDrawText(this,xLabel);
				break;
			case xLabelEx:
				obj = new CDrawText(this,xLabelEx);
				break;
			case xLabelEx2:
				obj = new CDrawText(this,xLabelEx2);
				break;
			case xJunction:
				obj = new CDrawJunction(this);
				break;
			case xNoConnect:
				obj = new CDrawNoConnect(this);
				break;
			case xBusSlash:
				obj = new CDrawBusSlash(this);
				break;
			case xPower:
				obj = new CDrawPower(this);
				break;
			case xPin:
			case xPinEx:
				obj = new CDrawPin(this);
				break;
			case xLine:
				obj = new CDrawPolygon(this,xLine);
				break;
			case xLineEx:
				obj = new CDrawPolygon(this,xLineEx);
				break;
			case xLineEx2:
				obj = new CDrawPolygon(this,xLineEx2);
				break;
			case xPolygon:
				obj = new CDrawPolygon(this,xPolygon);
				break;
			case xDash:
				obj = new CDrawPolygon(this,xDash);
				break;
			case xText:		// The old text type
				obj = new CDrawText(this,xText);
				break;
			case xTextEx:
				obj = new CDrawText(this,xTextEx);
				break;
			case xTextEx2:
				obj = new CDrawText(this,xTextEx2);
				break;
			case xCircle:
			case xCircleEx:
			case xCircleEx2:
			case xCircleEx3:
				obj = new CDrawSquare(this,xCircleEx3);
				break;
			case xSquare:
			case xSquareEx:
			case xSquareEx2:
			case xSquareEx3:
				obj = new CDrawSquare(this,xSquareEx3);
				break;
			case xArc:
			case xArcEx:
			case xArcEx2:
				obj = new CDrawPolygon(this,static_cast<ObjType>(tp));
				break;
			case xMethod:		// The old method type
			case xMethodEx:
			case xMethodEx2:
			case xMethodEx3:
				obj = new CDrawMethod(this);
				break;
			case xTag:
				obj = new CDrawTag(this);
				break;
			case xMetaFile:
				obj = new CDrawMetaFile(this);
				break;
			default:
				obj = NULL;
				break;
		}
		
		if (obj!=NULL) 
		{
			// Provide special handling for old objects
			switch (tp) {
				case xSquare:
				case xSquareEx:
				case xCircle:
				case xCircleEx:
					((CDrawSquare *)obj)->OldLoad(theArchive,tp);
					break;
				case xCircleEx2:
				case xSquareEx2:
					((CDrawSquare *)obj)->OldLoad2(theArchive);
					break;
				case xMethod:
					((CDrawMethod *)obj)->OldLoad(theArchive);
					break;
				case xMethodEx:
					((CDrawMethod *)obj)->OldLoad2(theArchive);
					break;
				case xMethodEx2:
					((CDrawMethod *)obj)->OldLoad3(theArchive);
					break;
				case xPin:
					((CDrawPin *)obj)->OldLoad(theArchive);
					break;
				default:
					obj->Load(theArchive);
					break;
			}

			// Now add object to linked list
			drawing.insert( drawing.end(), obj );
		}
  	}

  }

  catch( CException *e) 
  {
		CString s;
		CString msg;
		e->GetErrorMessage( msg.GetBuffer(256), 256, NULL );
		msg.ReleaseBuffer();
		s.Format("Cannot load file.\r\n%s",
			msg );
		AfxMessageBox( s );
		e->Delete();
  }


  return drawing.size() > 0;
}




////// Load a design from a file, loaded design will be selected //////
BOOL CTinyCadDoc::ReadFileXML(CXMLReader &xml, BOOL Details, drawingCollection &drawing, BOOL AlreadyStarted)
{
  try
  {
	// Read in the first item...
	CString name;
	if (!AlreadyStarted)
	{
		xml.nextTag( name );
		if (name != "TinyCAD")
		{
			Message(IDS_ABORTVERSION,MB_ICONEXCLAMATION);
			return FALSE;
		}
	}

	BOOL ResetMerge = TRUE;

	xml.intoTag();

	while (	xml.nextTag( name ) )
	{
		// Save the old layer setting
		CDrawingObject *obj = NULL;

		if (name == "DETAILS" && Details)
		{
			xml.intoTag();
			CString name;
			while (xml.nextTag( name ))
			{
				if (name == "Size")
				{
					xml.getAttribute( "width", MaxXSize );
					xml.getAttribute( "height", MaxYSize );
				}
				else if (name == "TITLE")
				{
					xml.getChildData( Title );
				}
				else if (name == "AUTHOR")
				{
					xml.getChildData( Author );
				}
				else if (name == "REVISION")
				{
					xml.getChildData( Revision );
				}
				else if (name == "DOCNUMBER")
				{
					xml.getChildData( DocNumber );
				}
				else if (name == "ORGANISATION")
				{
					xml.getChildData( Organisation );
				}
				else if (name == "SHEETS")
				{	
					xml.getChildData( Sheets );
				}
				else if (name == "SHOWS")
				{
					BYTE Shows = 0;
					xml.getChildData( Shows ); 
					ShowRulers = Shows & 2;
					ShowDetails = Shows & 1;
				}
				else if (name == "DATE")
				{
					xml.getChildData( DateStr );
				}
				else if (name == "GRID")
				{
					m_snap.LoadXML( xml );
				}
			}
		}
		else if (name == "FONT")
		{
			if (ResetMerge)
			{
				theOptions.ResetMerge();
				ResetMerge = FALSE;
			}
			theOptions.LoadFontXML(xml);
		}
		else if (name == "STYLE")
		{
			if (ResetMerge)
			{
				theOptions.ResetMerge();
				ResetMerge = FALSE;
			}
			theOptions.LoadStyleXML(xml);
		}
		else if (name == "FILL")
		{
			if (ResetMerge)
			{
				theOptions.ResetMerge();
				ResetMerge = FALSE;
			}
			theOptions.LoadFillStyleXML(xml);
		}
		else if (name == "META")
		{
			if (ResetMerge)
			{
				theOptions.ResetMerge();
				ResetMerge = FALSE;
			}
			theOptions.LoadMetaFileXML(xml);
		}
		else if (name == "SYMBOLDEF")
		{
			if (ResetMerge)
			{
				theOptions.ResetMerge();
				ResetMerge = FALSE;
			}
			theOptions.LoadSymbolXML(xml);
		}
		else if (name == "OPTIONS")
		{
			theOptions.LoadOptionsXML( xml, Details );
		}
		else if (name == CDrawBusSlash::GetXMLTag())
		{
			obj = new CDrawBusSlash(this);
		}
		else if (name == CDrawTag::GetXMLTag())
		{
			obj = new CDrawTag(this);
		}
		else if (name == CDrawJunction::GetXMLTag())
		{
			obj = new CDrawJunction(this);
		}
		else if (name == CDrawLine::GetXMLTag(xBus))
		{
			obj = new CDrawLine(this, xBus);
		}
		else if (name == CDrawLine::GetXMLTag(xWire))
		{
			obj = new CDrawLine(this, xWire);
		}
		else if (name == CDrawLine::GetXMLTag(xLineEx2))
		{
			obj = new CDrawLine(this, xLineEx2);
		}
		else if (name == CDrawMetaFile::GetXMLTag())
		{
			obj = new CDrawMetaFile(this);
		}
		else if (name == CDrawMethod::GetXMLTag())
		{
			obj = new CDrawMethod(this);
		}
		else if (name == CDrawNoConnect::GetXMLTag())
		{
			obj = new CDrawNoConnect(this);
		}
		else if (name == CDrawPin::GetXMLTag())
		{
			obj = new CDrawPin(this);
		}
		else if (name == CDrawRuler::GetXMLTag())
		{
			obj = new CDrawRuler(this, FALSE);
		}
		else if (name == CDrawSquare::GetXMLTag( TRUE ))
		{
			obj = new CDrawSquare(this, xSquareEx3);
		}
		else if (name == CDrawSquare::GetXMLTag( FALSE ))
		{
			obj = new CDrawSquare(this, xCircleEx3);
		}
		else if (name == CDrawText::GetXMLTag(xTextEx2))
		{
			obj = new CDrawText(this, xTextEx2);
		}
		else if (name == CDrawText::GetXMLTag(xBusNameEx))
		{
			obj = new CDrawText(this, xBusNameEx);
		}
		else if (name == CDrawText::GetXMLTag(xLabelEx2))
		{
			obj = new CDrawText(this, xLabelEx2);
		}
		else if (name == CDrawPolygon::GetXMLTag())
		{
			obj = new CDrawPolygon(this);
		}
		else if (name == CDrawPower::GetXMLTag())
		{
			obj = new CDrawPower(this);
		}

		if (obj != NULL)
		{
			// Load this object...
			obj->LoadXML( xml );

			// Now add object to linked list
			drawing.insert( drawing.end(), obj );
		}
	}

	xml.outofTag();
  }
  catch( CException *e ) 
  {
	  e->ReportError();
	  e->Delete();
	  return FALSE;
  }


  return drawing.size() > 0;
}




