/*
	TinyCAD program for schematic capture
	Copyright 1994/1995/2002 Matt Pyne.

	This program is free software; you can redistribute it and/or
	modify it under the terms of the GNU Lesser General Public
	License as published by the Free Software Foundation; either
	version 2.1 of the License, or (at your option) any later version.

	This program is distributed in the hope that it will be useful,
	but WITHOUT ANY WARRANTY; without even the implied warranty of
	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
	Lesser General Public License for more details.

	You should have received a copy of the GNU Lesser General Public
	License along with this library; if not, write to the Free Software
	Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
*/


#ifndef __NET_H__
#define __NET_H__

#define NetComment	";"		// The symbol used for comments




// This structure is for a balanced tree of nodes
class CNetListNode
{
public:
	int		m_NetList;					// The netlist this node is a member of
	
	CString m_label;					// The name of this node
	CString m_reference;
	CString m_pin;

	CPoint	m_a;						// The point where this node is on a design
	CDrawingObject *m_parent;			// The object which generated this node
	CDrawMethod *m_pMethod;				// The method which the pin belongs to

	// The constructor for this node
	CNetListNode(CDrawingObject *parent,CPoint a)
	{
	  m_a = a;
	  m_parent = parent;
	  m_NetList = -1;
	}

	// The destructor for this node and tree below it
	~CNetListNode()
	{
	}
};


inline bool operator<( const CPoint &a, const CPoint &b ) 
{
	if (a.x == b.x)
	{
		return a.y < b.y;
	}
	else
	{
		return a.x < b.x;
	}
}


typedef std::map<CString,int>	pinCollection;

// This class is used for spice file generation
class CNetListSymbol
{
public:
	
	// The pointer to the symbol and it's data
	CDrawMethod*	m_pMethod;				

	// The map of pins to their netlist (or power line)
	pinCollection	m_pins;
};



typedef std::map<CPoint,int>		nodeCollection;
typedef std::vector<CNetListNode>	nodeVector;
typedef std::map<int,nodeVector>	netCollection;


// Here is the class that holds it all together...
class CNetList
{
protected:

	int	m_CurrentNet;
	int GetNewNet();

	int Add(CNetListNode&);

	nodeCollection		m_nodes;


	// Create netlist and output as a PCB file (TinyCAD v1.xx)
	void WriteNetListFileTinyCAD( CTinyCadDoc *pDesign, const char *filename );

	// Create netlist and output as a PCB file (PADS-PCB)
	void WriteNetListFilePADS( CTinyCadDoc *pDesign, const char *filename );

	// Create netlist and output as a Eagle PCB script
	void WriteNetListFileEagle( CTinyCadDoc *pDesign, const char *filename );

	// Create netlist and output as a Protel PCB script
	void WriteNetListFileProtel( CTinyCadDoc *pDesign, const char *filename );


public:

	netCollection		m_nets;

	CNetList();
	virtual ~CNetList();

	// Perform the work of making a netlist from this design...
	void MakeNet( CTinyCadDoc *pDesign );

	// Create netlist and output as a PCB file
	void WriteNetListFile( int type, CTinyCadDoc *pDesign, const char *filename );

	// Create netlist and output as a SPICE file
	void WriteSpiceFile( CTinyCadDoc *pDesign, const char *filename );
};


// The structure which defines which errors/warnings are in use
struct ErrorTest {
	int DupRef	: 1;		// Duplicated references 0
	int UnConnect	: 1;		// Unconnected items 1
	int NoConnect	: 1;		// Mode than one item on a no-connect net 2
	int Power	: 1;		// Power connected to power 3
	int OutputPwr	: 1;		// Power connected to an output 4
	int Output	: 1;		// Output connected to an output 5
	int NoOutput	: 1;		// No outputs driving inputs 6
	int UnConnected : 1;		// Unconnected nets 7
};

// The type of nets that exist
enum NetType 
{	
		nUnknown, 			// The net has not been evaluated yet						0
		nPassive,			// The net only contains passive connections				1
		nInput,				// The net only contains inputs								2
		nOutput,			// The net contains an output and some inputs/passives		3
		nBiDir,				// The net contains a bi-dir/inputs/passives/tri-state		4
		nPower,				// The net contains a power item and some inputs/passives	5
		nNoConnect,			// The net contains a no-connect and one other item			6
};



// The dialog for the electrical rules check box
class CDlgERCBox : public CDialog {
	ErrorTest 	theErrorTest;

public:	
 	CDlgERCBox(CWnd *pParentWnd = NULL) : CDialog( IDD_ERCBOX, pParentWnd ) {}
	BOOL OnInitDialog();
	virtual void OnOK();

	void SetErrorTest(ErrorTest NewErrorTest) { theErrorTest = NewErrorTest; }
	ErrorTest GetErrorTest() { return theErrorTest; }
};


#endif
