/*
	TinyCAD program for schematic capture
	Copyright 1994/1995/2002,2003 Matt Pyne.

	This program is free software; you can redistribute it and/or
	modify it under the terms of the GNU Lesser General Public
	License as published by the Free Software Foundation; either
	version 2.1 of the License, or (at your option) any later version.

	This program is distributed in the hope that it will be useful,
	but WITHOUT ANY WARRANTY; without even the implied warranty of
	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
	Lesser General Public License for more details.

	You should have received a copy of the GNU Lesser General Public
	License along with this library; if not, write to the Free Software
	Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
*/


#ifndef __OPTION_H__
#define __OPTION_H__


#include "Symbol.h"
#include "Colour.h"

#define fDEFAULT_FONT 	"Arial"
#define fDEFAULT_STYLE 	"Regular"

#define hRESOURCE		short

#define	hFONT			hRESOURCE					// The font handle type
#define hSTYLE			hRESOURCE					// The line style type
#define hFILL			hRESOURCE					// The fill style type
#define hMETAFILE		hRESOURCE
#define hSYMBOL			hRESOURCE

class CTinyCadDoc;

class Transform;
class TransformSnap;

class LineStyle {
public:
	LONG		Style;
	LONG		Thickness;
	COLORREF	Colour;
};


class FillStyle
{
public:
	int			Index;
	COLORREF	Colour;
};


////// The resource settings //////



class CDocResource {
private:
	hRESOURCE	ResourceNumber;					// The identifier for this resource
	hRESOURCE	MergeNumber;					// Used during merging of two resource tables
	BOOL		InUse;							// Used during Garbage collection

public:

	CDocResource 	*next;							// The Next font in this list

	class LayerBox;
	friend 	LayerBox;


	// The constructor
	CDocResource();

	// The virtual destructor
	virtual ~CDocResource();

	// Convert an old resource number into a new resource number
	hRESOURCE ConvertOld(hRESOURCE n);	
	// Add a new resource to this list
	hRESOURCE Add(CDocResource *,hRESOURCE = -1);
	// Get a resource from the list
	CDocResource *Get(hRESOURCE);

	// The Garbage collection facilities
	void Tag(hRESOURCE);
	void UnTag();

	// Clear the merge numbers
	void ResetMerge();

	void SaveXML(CTinyCadDoc *pDesign, const char *name, CXMLWriter &xml);
	void Load(CTinyCadDoc *pDesign, CStream &);

	hRESOURCE	GetResourceNumber() { return ResourceNumber; }

	// The virtual parts of the resource
	virtual void SaveItemXML( CTinyCadDoc *pDesign, CXMLWriter &xml) = 0;
	virtual void LoadItem(CTinyCadDoc *pDesign, CStream &, hRESOURCE) = 0;
	virtual void LoadItemXML(CTinyCadDoc *pDesign, CXMLReader &xml) = 0;
	virtual	BOOL Compare(CDocResource *) = 0;			// Compare this resource with another resource
};



// Storage of font settings	
class ListOfFonts : public CDocResource {
private:
	LOGFONT	FontStyle;							// The Font this describes
public:
	// The constructor
	ListOfFonts(LOGFONT f);

	virtual ~ListOfFonts()
	{
	}

	// Get the data associated with this resource
	LOGFONT *GetFontStyle() { return &FontStyle; }

	// Compare this resource with another resource of the same type
	BOOL Compare(CDocResource *);
	virtual void SaveItemXML( CTinyCadDoc *pDesign, CXMLWriter &xml);
	virtual void LoadItemXML(CTinyCadDoc *pDesign, CXMLReader &xml);
	void LoadItem(CTinyCadDoc *pDesign, CStream &, hRESOURCE);
};



// Storage of line style settings	
class ListOfStyles : public CDocResource {
private:
	LineStyle	theLineStyle;						// The Line style this describes
public:
	// The constructor
	ListOfStyles(LineStyle s) 
	{ 
		theLineStyle = s; 
	}

	virtual ~ListOfStyles()
	{
	}


	// Get the data associated with this resource
	LineStyle *GetLineStyle() { return &theLineStyle; }

	// Compare this resource with another resource of the same type
	BOOL Compare(CDocResource *);
	virtual void SaveItemXML( CTinyCadDoc *pDesign, CXMLWriter &xml);
	void LoadItem(CTinyCadDoc *pDesign, CStream &, hRESOURCE);
	virtual void LoadItemXML(CTinyCadDoc *pDesign, CXMLReader &xml);

};


// Storage of fill style settings	
class ListOfFillStyles : public CDocResource {
private:
	FillStyle	theFillStyle;						// The Fill style this describes
public:
	// The constructor
	ListOfFillStyles(FillStyle s) 
	{ 
		theFillStyle = s; 
	}

	virtual ~ListOfFillStyles()
	{
	}


	// Get the data associated with this resource
	FillStyle *GetFillStyle() { return &theFillStyle; }

	// Compare this resource with another resource of the same type
	BOOL Compare(CDocResource *);
	virtual void SaveItemXML( CTinyCadDoc *pDesign, CXMLWriter &xml);
	void LoadItem(CTinyCadDoc *pDesign, CStream &, hRESOURCE);
	virtual void LoadItemXML(CTinyCadDoc *pDesign, CXMLReader &xml);
};



class ListOfMetaFiles : public CDocResource {
	HENHMETAFILE theMetaFile;
public:
	// The constructor
	ListOfMetaFiles(HENHMETAFILE s)
	{ 
		theMetaFile = s; 
	}

	virtual ~ListOfMetaFiles()
	{
	}


	// Get the data associated with this resource
	HENHMETAFILE GetMetaFile() { return theMetaFile; }

	// Compare this resource with another resource of the same type
	BOOL Compare(CDocResource *);
	virtual void SaveItemXML( CTinyCadDoc *pDesign, CXMLWriter &xml);
	void LoadItem(CTinyCadDoc *pDesign, CStream &, hRESOURCE);
	virtual void LoadItemXML(CTinyCadDoc *pDesign, CXMLReader &xml);
};

class ListOfSymbols : public CDocResource 
{
	CDesignFileSymbol	*theSymbol;
public:
	// The constructor
	ListOfSymbols(CDesignFileSymbol *s) : CDocResource() 
	{ 
		theSymbol = s; 
	}

	// The destructor
	virtual ~ListOfSymbols();

	// Get the data associated with this resource
	CDesignFileSymbol* GetSymbol() { return theSymbol; }

	// Compare this resource with another resource of the same type
	BOOL Compare(CDocResource *);
	virtual void SaveItemXML( CTinyCadDoc *pDesign, CXMLWriter &xml);
	void LoadItem(CTinyCadDoc *pDesign, CStream &, hRESOURCE);
	virtual void LoadItemXML(CTinyCadDoc *pDesign, CXMLReader &xml);


	// Reload from the library
	void ResetSymbol();
};



class Option 
{
	CTinyCadDoc *m_pDesign;

	ListOfFonts	 *theFontList;
	ListOfStyles *theStyleList;
	ListOfFillStyles *theFillStyleList;
	ListOfMetaFiles *theMetaFileList;
	ListOfSymbols *theSymbolList;

	// The current settings
	hFONT	CurrentFont;						// The default font number
	hSTYLE	CurrentStyle;						// The default line style
	hFILL	CurrentFillStyle;					// The default line style

	// The current grid settings
	BOOL	GridShow;							// Should we show the grid?
	LONG	Units;								// The units to use
	LONG	ScrollBar;							// The scroll bar increments

	// Some default settings for dialogs
	int		PinLength;							// The default pin length
	int		PinNumberPos;						// The default pin number position

	// The current auto-snap settings
	BOOL	AutoSnap;
	BOOL	AutoDrag;
	BOOL	AutoJunc;
	int		AutoSnapRange;

	// The user defined colours
	static const char *ColourNames[ 7 ];
	static const COLORREF ColoursDefs[7];
	COLORREF	Colours[7];


public:

	Option();									// The Constructor
	~Option();

	void Init( CTinyCadDoc *pDesign );


	// Untag all of our resources
	void UnTag();

	////// The font operators ////

	// Add a font to the list and return it's new index number
	hFONT AddFont(LOGFONT *lf,hFONT MergeNumber = -1) { return theFontList->Add(new ListOfFonts(*lf), MergeNumber); }

	// Load the font table from a file
	void LoadFonts(CStream &theFile) { theFontList->Load(m_pDesign,theFile); }
	// Convert an old font number into a new font number
	hFONT GetNewFontNumber(hFONT n) { return theFontList->ConvertOld(n); }
	// Get a font from the number
	LOGFONT *GetFont(hFONT n)  { return ((ListOfFonts *)theFontList->Get(n))->GetFontStyle(); }
	hFONT ChangeFontSize(hFONT n, int delta_height, int width );
	// Specify this font is in use
	void TagFont(hFONT n) { theFontList->Tag(n); }
	// Bring up the user change font box
	void ChooseFont(CDC &,HWND = NULL);	
	hFONT  GetCurrentFont() { return CurrentFont; }
	void SetCurrentFont(hFONT n) { CurrentFont = n; }
	int GetFontWidth( hFONT n );


	////// The Line Style operators //////

	// Add a line style to the list and return it's new index number
	hSTYLE AddStyle(LineStyle *ls,hFONT MergeNumber = -1) { return theStyleList->Add(new ListOfStyles(*ls), MergeNumber); }
	// Load the line style table from a file
	void LoadStyles(CStream &theFile) { theStyleList->Load(m_pDesign,theFile); }
	// Convert an old line style number into a new font number
	hSTYLE GetNewStyleNumber(hSTYLE n) { return theStyleList->ConvertOld(n); }
	// Get a line style from the number
	LineStyle *GetStyle(hSTYLE n)  { return ((ListOfStyles *)theStyleList->Get(n))->GetLineStyle(); }
	// Specify this line style is in use
	void TagStyle(hSTYLE n) { theStyleList->Tag(n); }
	hSTYLE  GetCurrentStyle() { return CurrentStyle; }
	void SetCurrentStyle(hSTYLE n) { CurrentStyle = n; }


	////// The File Style operators //////

	// Add a fill style to the list and return it's new index number
	hFILL AddFillStyle(FillStyle *ls,hFONT MergeNumber = -1) { return theFillStyleList->Add(new ListOfFillStyles(*ls), MergeNumber); }
	// Load the fill style table from a file
	void LoadFillStyles(CStream &theFile) { theFillStyleList->Load(m_pDesign,theFile); }
	// Convert an old line style number into a new font number
	hFILL GetNewFillStyleNumber(hFILL n) { return theFillStyleList->ConvertOld(n); }
	// Get a fill style from the number
	FillStyle *GetFillStyle(hSTYLE n)  { return ((ListOfFillStyles *)theFillStyleList->Get(n))->GetFillStyle(); }
	// Specify this fill style is in use
	void TagFillStyle(hSTYLE n) { theFillStyleList->Tag(n); }
	hFILL  GetCurrentFillStyle() { return CurrentFillStyle; }
	void SetCurrentFillStyle(hFILL n) { CurrentFillStyle = n; }


	////// The Meta File operators //////

	// Add a line style to the list and return it's new index number
	hMETAFILE AddMetaFile(HENHMETAFILE s,hMETAFILE MergeNumber = -1) { return theMetaFileList->Add(new ListOfMetaFiles(s), MergeNumber); }
	// Load the line style table from a file
	void LoadMetaFiles(CStream &theFile) { theMetaFileList->Load(m_pDesign,theFile); }
	// Convert an old line style number into a new font number
	hMETAFILE GetNewMetaFileNumber(hMETAFILE n) { return theMetaFileList->ConvertOld(n); }
	// Get a line style from the number
	HENHMETAFILE GetMetaFile(hMETAFILE n)  { return ((ListOfMetaFiles *)theMetaFileList->Get(n))->GetMetaFile(); }
	// Specify this line style is in use
	void TagMetaFile(hSTYLE n) { theMetaFileList->Tag(n); }



	////// The Symbols operators //////

	// Add a line style to the list and return it's new index number
	hSYMBOL AddSymbol(CDesignFileSymbol *s,hSYMBOL MergeNumber = -1) { return theSymbolList->Add(new ListOfSymbols(s), MergeNumber); }
	// Load the line style table from a file
	void LoadSymbols(CStream &theFile) { theSymbolList->Load(m_pDesign,theFile); }
	// Convert an old line style number into a new font number
	hSYMBOL GetNewSymbolNumber(hSYMBOL n) { return theSymbolList->ConvertOld(n); }
	// Get a line style from the number
	CDesignFileSymbol *GetSymbol(hSYMBOL n)  { return ((ListOfSymbols *)theSymbolList->Get(n))->GetSymbol(); }
	// Specify this line style is in use
	void TagSymbol(hSYMBOL n) { theSymbolList->Tag(n); }
	// Reload the symbols from their libraries
	void ResetSymbols();



	////// Save the current settings along with the design //////
	void SaveOptions( CStream &theArchive );
	void LoadOptions( CStream &theArchive, BOOL keep );


	////// Get the user defined colours //////
	void ResetColours();
	COLORREF GetColour( user_colour u );
	void SetColour( user_colour u, COLORREF c );


	////// The XML operations //////

	void SaveFontsXML(CXMLWriter &xml) { theFontList->SaveXML(m_pDesign,"FONT", xml); }
	void SaveStylesXML(CXMLWriter &xml) { theStyleList->SaveXML(m_pDesign,"STYLE", xml); }
	void SaveFillStylesXML(CXMLWriter &xml) { theFillStyleList->SaveXML(m_pDesign,"FILL", xml); }
	void SaveMetaFilesXML(CXMLWriter &xml) { theMetaFileList->SaveXML(m_pDesign,"META", xml); }
	void SaveSymbolsXML(CXMLWriter &xml) { theSymbolList->SaveXML(m_pDesign,"SYMBOLDEF", xml); }

	void SaveOptionsXML(CXMLWriter &xml);

	void LoadFontXML(CXMLReader &xml) { theFontList->LoadItemXML(m_pDesign,xml); }
	void LoadStyleXML(CXMLReader &xml) { theStyleList->LoadItemXML(m_pDesign,xml); }
	void LoadFillStyleXML(CXMLReader &xml) { theFillStyleList->LoadItemXML(m_pDesign,xml); }
	void LoadMetaFileXML(CXMLReader &xml) { theMetaFileList->LoadItemXML(m_pDesign,xml); }
	void LoadSymbolXML(CXMLReader &xml) { theSymbolList->LoadItemXML(m_pDesign,xml); }

	void LoadOptionsXML(CXMLReader &xml, BOOL keep);

	void ResetMerge()
	{
		theFontList->ResetMerge();
		theStyleList->ResetMerge();
		theFillStyleList->ResetMerge();
		theMetaFileList->ResetMerge();
		theSymbolList->ResetMerge();
	}


	////// The grid operators //////

	BOOL ShowGrid() { return GridShow; }		// The size of the current grid
	void SetGridShow( BOOL s ) { GridShow = s; }
	int	 GetUnits() { return Units; }
	void SetUnits( int u ) { Units = u; }
	int  GetScrollBar() { return ScrollBar; }
	void SetScrollBar( int s ) { ScrollBar = s; }
	CString PointToUnit(CPoint);
	CString PointToDisplay(CPoint,BOOL horiz);

	////// The auto-wire settins //////
	BOOL GetAutoSnap() { return AutoSnap; }
	BOOL GetAutoDrag() { return AutoDrag; }
	BOOL GetAutoJunc() { return AutoJunc; }
	int	 GetAutoSnapRange() { return AutoSnapRange; }

	void SetAutoSnap(BOOL r) { AutoSnap = r; }
	void SetAutoDrag(BOOL r) { AutoDrag = r; }
	void SetAutoJunc(BOOL r) { AutoJunc = r; }
	void SetAutoSnapRange( int r ) { AutoSnapRange = r; }


	////// Some default settings for dialogs ////// 

	int		GetPinLength() { return PinLength; }
	void	SetPinLength( int pl );

	int		GetPinNumberPos() { return PinNumberPos; }
	void	SetPinNumberPos( int pl );

};




#endif
