/*
	TinyCAD program for schematic capture
	Copyright 1994/1995/2002,2003 Matt Pyne.

	This program is free software; you can redistribute it and/or
	modify it under the terms of the GNU Lesser General Public
	License as published by the Free Software Foundation; either
	version 2.1 of the License, or (at your option) any later version.

	This program is distributed in the hope that it will be useful,
	but WITHOUT ANY WARRANTY; without even the implied warranty of
	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
	Lesser General Public License for more details.

	You should have received a copy of the GNU Lesser General Public
	License along with this library; if not, write to the Free Software
	Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
*/



// This handles the actual drawing of objects
#include "stdafx.h"
#include <math.h>
#include "TinyCadView.h"
#include "diag.h"
#include "colour.h"
#include "option.h"
#include "LineUtils.h"

#define OUTLINE_SPACER 5


////// The Polygon Class //////


CDrawPolygon::CDrawPolygon(CTinyCadDoc *pDesign, ObjType NewType)
: CDrawingObject( pDesign )
{
	m_segment=1;
	xtype=NewType;
	m_re_edit = FALSE;

	switch (xtype) 
	{
	case xPolygon:
	case xLineEx2:
		Style = m_pDesign->GetOptions()->GetCurrentStyle();
		Fill = fsNONE;
		break;
	case xDash:
		Style = fDASH;
		Fill = fsNONE;
		break;
	default:
		Style = fLINE;
		Fill = fsNONE;
	}
}



void CDrawPolygon::TagResources()
{
   m_pDesign->GetOptions()->TagStyle(Style);
   m_pDesign->GetOptions()->TagFillStyle(Fill);
}

const char* CDrawPolygon::GetXMLTag()
{
	return "POLYGON";
}

// Load and save to an XML file
void CDrawPolygon::SaveXML( CXMLWriter &xml )
{
	xml.addTag(GetXMLTag());

	xml.addAttribute( "pos", CDPoint(a) );
	xml.addAttribute( "style", Style );
	xml.addAttribute( "fill", Fill );

	DWORD lines = m_handles.size();

	arcpointCollection::iterator it = m_handles.begin();
	while (it != m_handles.end())
	{
		xml.addTag("POINT");
		xml.addAttribute( "pos", 
			CDPoint( CDPoint::makeXMLUnits((*it).x), CDPoint::makeXMLUnits((*it).y) ) );
		xml.addAttribute( "arc", (*it).arc );
		xml.closeTag();
		++ it;
	}

	xml.closeTag();
}

void CDrawPolygon::LoadXML( CXMLReader &xml )
{
	xml.getAttribute( "pos", a );
	xml.getAttribute( "style", Style );
	xml.getAttribute( "fill", Fill );


	m_handles.erase( m_handles.begin(), m_handles.end() );

	xml.intoTag();

	CString name;
	while (xml.nextTag(name))
	{
		if (name == "POINT")
		{
			CDPoint p;
			int arc = 0;
			xml.getAttribute( "pos", p );
			xml.getAttribute( "arc", arc );

			CArcPoint q;
			q.x = p.unmakeXMLUnits( p.x );
			q.y = p.unmakeXMLUnits( p.y );
			q.arc = static_cast<CArcPoint::arc_type>(arc);
			m_handles.push_back( q );
		}
	}

	xml.outofTag();

	Style = m_pDesign->GetOptions()->GetNewStyleNumber(Style);
	Fill = m_pDesign->GetOptions()->GetNewFillStyleNumber(Fill);

	xtype = xArcEx2;

	CalcBoundingRect();
}


// Load the line from a file
void CDrawPolygon::Load(CStream& archive )
{
	m_handles.erase( m_handles.begin(), m_handles.end() );

	if (xtype != xPolygon)
	{
	 	a = ReadPoint(archive);
   		b = ReadPoint(archive);
		m_handles.push_back(CArcPoint(0,0));

		CPoint p( b - a );
		if (xtype == xArcEx || xtype == xArcEx2)
		{
			m_handles.push_back(CArcPoint(p.x,p.y, CArcPoint::Arc_out));
		}
		else
		{
			m_handles.push_back(CArcPoint(p));
		}
	}

	switch (xtype) 
	{
		case xDash:
		case xLine:
			// Premote this type
			xtype = xLineEx2;
			break;
		case xArcEx:
		case xLineEx:
			// Load the line thickness
			BYTE LineThickness;
			archive >> LineThickness;
			LineStyle l;
			l.Colour = cBLACK;
			l.Style = PS_SOLID;
			l.Thickness = LineThickness;
			Style = m_pDesign->GetOptions()->AddStyle(&l);
			Style = m_pDesign->GetOptions()->GetNewStyleNumber(Style);
			Fill = fsNONE;
			if (xtype == xLine)
			{
				xtype = xLineEx2;
			}
			if (xtype == xArcEx)
			{
				xtype = xArcEx2;
			}
			break;
		case xArcEx2:
		case xLineEx2:
			archive >> Style;
			Fill = fsNONE;
			Style = m_pDesign->GetOptions()->GetNewStyleNumber(Style);
			break;
		case xPolygon:
			// Write out the points for this polygon
			{
				DWORD version;
				DWORD lines;  	
				archive >> version;
				archive >> lines;
				a = ReadPoint(archive);
				while (lines > 0)
				{
					CArcPoint q;
					if (version == 0)
					{
						q = ReadPoint( archive );
					}
					else
					{
						q.ReadPoint( archive );
					}
					m_handles.push_back( q );
					lines --;
				}
				archive >> Style;
				archive >> Fill;  	
				Style = m_pDesign->GetOptions()->GetNewStyleNumber(Style);
				Fill = m_pDesign->GetOptions()->GetNewFillStyleNumber(Fill);

			}
		case xArc:
			// Premote this type
			xtype = xArcEx2;
			break;
	}
	
	CalcBoundingRect();
}


// Rotate this object about a point
void CDrawPolygon::Rotate(CPoint p,int dir)
{
  // Now rotate each point around the rotational point...
	arcpointCollection::iterator it = m_handles.begin();
	while (it != m_handles.end())
	{
		CArcPoint qp(*it + a);
		CArcPoint::arc_type at = (*it).arc;

		// Translate this point so the rotational point is the origin
		qp = CPoint(qp.x-p.x,qp.y-p.y);

		// Perfrom the rotation
		switch (dir) {
		case 2: // Left
			qp = CPoint(qp.y,-qp.x);
			break;		
		case 3: // Right
			qp = CPoint(-qp.y,qp.x);
			break;
		case 4: // Mirror
			qp = CPoint(-qp.x,qp.y);
			break;
		}

		if (dir != 4)
		{
			switch (at)
			{
			case CArcPoint::Arc_none:
				break;
			case CArcPoint::Arc_in:
				at = CArcPoint::Arc_out;
				break;
			case CArcPoint::Arc_out:
				at = CArcPoint::Arc_in;
				break;
			}
		}

		// Re-translate the points back to the original location
		*it = CArcPoint(qp.x+p.x,qp.y+p.y,at) - a;

		++ it;
	}

	// Now re-calc the bounding rectangle
	CalcBoundingRect();
}


// Re-calculate bounding rect
// Used when the polygon is inside a symbol.
//
void CDrawPolygon::CalcBoundingRect()
{
	// First flattern out the arcs
	FlatternPath();

	// Now Calc the bounding rect
	CRect r;

	if (m_points.size() >0)
	{
		CDPoint q = m_points.front() + a;
		r = CRect(q.x,q.y,q.x,q.y);
	}
	else
	{
		r = CRect(a.x,a.y,a.x,a.y);
	}

	// Find the bounding rectangle
	pointCollection::iterator it = m_points.begin();
	while (it != m_points.end())
	{
		CDPoint qp = *it + a;

		r.left = min(r.left,qp.x);
		r.top = min(r.top,qp.y);
		r.right = max(r.right,qp.x);
		r.bottom = max(r.bottom,qp.y);

		++ it;
	}

	// Now move everything so that a and b
	// lie on the bounding rectangle

	CPoint diff_a = CPoint(r.left,r.top) - a;
	arcpointCollection::iterator itx = m_handles.begin();
	while (itx != m_handles.end())
	{
		*itx = *itx - diff_a;
		++ itx;
	}

	a = CPoint(r.left,r.top);
	b = CPoint(r.right,r.bottom);

	FlatternPath();
}


void CDrawPolygon::ToAngle()
{
  int FromX,FromY,FromC;
  
  CArcPoint last = a;
  if (m_handles.size() > 0)
  {
	last = m_handles.back() + a;
  }

  switch (g_EditToolBar.m_LineEdit.GetAngle()) 
  {
	case LINEBOX_90:
		if (last.x==b.x)
			g_EditToolBar.m_LineEdit.mode=2;
		if (last.y==b.y)
			g_EditToolBar.m_LineEdit.mode=1;
		if (g_EditToolBar.m_LineEdit.mode==0)
			g_EditToolBar.m_LineEdit.mode=1;
		break;
	case LINEBOX_45:
		g_EditToolBar.m_LineEdit.mode=0;
		// Which 45degree slope is it closest to?
		FromX=abs(last.x-b.x);
		FromY=abs(last.y-b.y);
		FromC=abs(FromX-FromY);
		// Is it closest to the X axis?
		if (FromX<FromY && FromX<FromC)
			b.x=last.x;
		// Is it closest to the Y axis?
		else if (FromY<FromC)
			b.y=last.y;
		// Must be closest to the 45 degree line
		else
			b.x=last.x+(b.x>last.x ? FromY : -FromY);
		break;
	default:
		g_EditToolBar.m_LineEdit.mode=0;
  }
}


BOOL CDrawPolygon::RButtonDown(CPoint p, CPoint s)
{
	// Get the current location of the mouse
	CPoint mp;
	GetCursorPos( &mp );

	// Now bring up the context menu..
	CMenu menu;
	menu.LoadMenu( IDR_POLYGON );
	menu.GetSubMenu(0)->TrackPopupMenu( TPM_LEFTALIGN|TPM_RIGHTBUTTON,
		mp.x,mp.y, AfxGetMainWnd(), NULL );
	
	return TRUE;
}

void CDrawPolygon::FindNearestSegment( CPoint p, int &line, int &handle )
{
	double distance = 1e10;

	int s = 0;
	int h = 0;
	CDPoint dla = m_points.back();
	CPoint la( dla.x, dla.y );
	pointCollection::iterator it = m_points.begin();
	arcpointCollection::iterator itz = m_handles.begin();
	while (it != m_points.end())
	{
		CDPoint a_np = *it + a;
		CPoint np(a_np.x, a_np.y);
		CLineUtils l( la, np);
		CPoint loc_i;
		double distance_i;

		distance_i = l.DistanceFromPoint( p, loc_i );
		if (distance_i <= distance)
		{
			distance = distance_i;
			line = h;
		}

		++ s;

		if (s == (*itz).flatterned_segment)
		{
			++ h;	
			++ itz;
		}

		la = np;
		++ it;
	}

	// Now find the nearest handle..
	distance = 1e10;
	s = 0;
	itz = m_handles.begin();
	while (itz != m_handles.end())
	{
		CArcPoint qp = *itz + a;
		double distance_i = 
			(qp.x-p.x)*(qp.x-p.x) + (qp.y-p.y)*(qp.y-p.y);
		if (distance_i < distance)
		{
			distance = distance_i;
			handle = s;
		}

		++ s;
		++ itz;
	}
}



// For the context menu
void CDrawPolygon::ContextMenu( CPoint p, UINT id )
{
	if (!m_segment)
	{
		switch (id)
		{
		case ID_CONTEXT_ARCIN:
			g_EditToolBar.m_LineEdit.SetArcType( CArcPoint::Arc_in );
			break;
		case ID_CONTEXT_ARCOUT:
			g_EditToolBar.m_LineEdit.SetArcType( CArcPoint::Arc_out );
			break;
		case ID_CONTEXT_FREELINE:
			g_EditToolBar.m_LineEdit.SetArcType( CArcPoint::Arc_none );
			break;
		}
	}
	else
	{
		// Find the nearest line m_segment...
		// and locate the nearest handle to it...
		int line;
		int handle;

		Display();
		FindNearestSegment( p, line, handle );

		switch (id)
		{
		case ID_CONTEXT_ARCIN:
			if (line > 0)
			{
				m_handles[ line ].arc = CArcPoint::Arc_in;
			}
			break;
		case ID_CONTEXT_ARCOUT:
			if (line > 0)
			{
				m_handles[ line ].arc = CArcPoint::Arc_out;
			}
			break;
		case ID_CONTEXT_FREELINE:
			if (line > 0)
			{
				m_handles[ line ].arc = CArcPoint::Arc_none;
			}
			break;
		case ID_CONTEXT_DELETEHANDLE:
			if (m_handles.size() > 2)
			{
				m_handles.erase( m_handles.begin() + handle );
			}
			break;
		case ID_CONTEXT_ADDHANDLE:
			m_handles.insert( m_handles.begin() + line, CArcPoint( p - a ) );
			break;
		}

		CalcBoundingRect();
		Display();
	}
}

int  CDrawPolygon::GetContextMenu()
{
	return IDR_POLYGON_EDIT;
}



// For the context menu
void CDrawPolygon::FinishDrawing( CPoint p )
{
	// Ok, we can place
	if (!m_segment)
	{
		// First line
		LButtonDown(p,p);
	}
	// Second line..
	LButtonDown(p,p);

	// Now Store...
	CDrawingObject *obj = Store();

	// .. and clear out...
	m_handles.erase( m_handles.begin(), m_handles.end() );
	m_points.erase( m_points.begin(), m_points.end() );

	// Now select the top object and switch back
	// to the Edit tool
	m_pDesign->UnSelect();
	m_pDesign->Select( obj );
	m_pDesign->SelectObject( new CDrawEditItem(m_pDesign) );
}



void CDrawPolygon::DblLButtonDown(CPoint p,CPoint)
{
	FinishDrawing( p );
}


void CDrawPolygon::LButtonUp(CPoint p)
{
	// Can't drag out a polygon!
}


void CDrawPolygon::NewOptions()
{
  	Display();
	Style = m_pDesign->GetOptions()->GetCurrentStyle();
	Fill =  m_pDesign->GetOptions()->GetCurrentFillStyle();
	Display();
}


CString CDrawPolygon::GetName() const
{
	return "Polygon";
}

int CDrawPolygon::getMenuID() 
{ 
	return IDM_TOOLPOLYGON;
}


void CDrawPolygon::LButtonDown(CPoint p, CPoint)
{
  // New undo level for each placement...
  m_pDesign->BeginNewChangeSet();

  if (m_segment) 
  {
	a=p;
	b=p;
	m_segment=!m_segment;
	m_handles.push_back( CPoint(0,0) );
  } 
  else 
  {
	b=p;
	ToAngle();

	// Remove from screen
	Display();
	CArcPoint last = a;
	if (m_handles.size() > 0)
	{
		last = m_handles.back() + a;
	}

	CArcPoint c=b;

	switch (g_EditToolBar.m_LineEdit.mode) 
	{
		case 1: c=CPoint(b.x,last.y);
			break;
		case 2: c=CPoint(last.x,b.y);
			break;
	}

	c = c - a;

	CArcPoint p( c.x, c.y, g_EditToolBar.m_LineEdit.GetArcType());

	// Only push back if this point is different from
	// the last
	if (m_handles.size() > 0)
	{
		if (p != m_handles.back())
		{
			m_handles.push_back( p );
		}
	}
	else
	{
		m_handles.push_back( p );
	}

	FlatternPath();
	Display();	// Write to screen
	m_segment=0;

	ToAngle();
  }

  Display();
}


void CDrawPolygon::Move(CPoint p)
{
  if (!m_segment) 
  {
	Display();

	b=p;
	FlatternPath();
	ToAngle();

	Display();
  }

}





ObjType CDrawPolygon::GetType()
{
  return xPolygon;
}

void CDrawPolygon::Display( BOOL erase )
{
	LineStyle *theStyle = m_pDesign->GetOptions()->GetStyle(Style);
	int width = max(2,theStyle->Thickness) * 2;

  CRect r(min(a.x,b.x),min(a.y,b.y),
	  max(a.x,b.x),max(a.y,b.y));


	pointCollection::iterator it = m_points.begin();
	while (it != m_points.end())
	{
		CDPoint qp = *it + a;
		  r.top = min(r.top, qp.y);
		  r.left = min(r.left, qp.x);
		  r.bottom = max(r.bottom, qp.y);
		  r.right = max(r.right, qp.x);

		++ it;
	}

	m_pDesign->InvalidateRect( r, erase, OUTLINE_SPACER * 2 );
}

void CDrawPolygon::PaintHandles( Context &dc )
{
	// Put some handles around this object
	arcpointCollection::iterator it = m_handles.begin();
	dc.SelectBrush(cBLACK);
	dc.SelectPen(PS_SOLID,0,cBLACK);

	while (it != m_handles.end())
	{
		CArcPoint p = *it + a;
		CRect r(p.x-2,p.y-2,p.x+2,p.y+2);
		dc.Rectangle(r);

		++ it;
	}

	// Now draw the tracker for resizing...
	CalcBoundingRect();
	CRect r(a.x - OUTLINE_SPACER,a.y - OUTLINE_SPACER,
		b.x + OUTLINE_SPACER,b.y + OUTLINE_SPACER);

	CRectTracker tracker( dc.theTransform.Scale(r), CRectTracker::dottedLine | CRectTracker::resizeOutside  );
	tracker.Draw( dc.dc );
}


// Move fields of this object about
int CDrawPolygon::IsInsideField(CPoint p)
{
	// Check the handles around this object
	int s = 0;
	arcpointCollection::iterator it = m_handles.begin();
	while (it != m_handles.end())
	{
		CArcPoint qp = *it + a;
		CRect r(qp.x-2,qp.y-2,qp.x+2,qp.y+2);
		if (r.PtInRect(p))
		{
			return s + 1000;
		}

		++ s;
		++ it;
	}

	// Now is it inside our tracker?
	CRect r(a.x - OUTLINE_SPACER,a.y - OUTLINE_SPACER,
		b.x + OUTLINE_SPACER,b.y + OUTLINE_SPACER);

	CRectTracker tracker( r, CRectTracker::dottedLine | CRectTracker::resizeOutside  );
	s = tracker.HitTest( p );

	if (s == 8)
	{
		s = 11;
	}

	return s;
}

void CDrawPolygon::MoveField(int w, CPoint r)
{
	Display();

	if (w >= 1000)
	{
		// Individual handle movement...
		m_handles[w - 1000] = m_handles[w - 1000] + r;
	}
	else
	{
		CRect rect( a.x, a.y, b.x, b.y );

		// Tracker movement...
		switch (w)
		{
		case CRectTracker::hitTopLeft:
			rect.left += r.x;
			rect.top += r.y;
			break;
		case CRectTracker::hitTopRight:
			rect.right += r.x;
			rect.top += r.y;
			break;
		case CRectTracker::hitBottomRight:
			rect.right += r.x;
			rect.bottom += r.y;
			break;
		case CRectTracker::hitBottomLeft:
			rect.left += r.x;
			rect.bottom += r.y;
			break;	
		case CRectTracker::hitTop:
			rect.top += r.y;
			break;		
		case CRectTracker::hitRight:
			rect.right += r.x;
			break;		
		case CRectTracker::hitBottom:
			rect.bottom += r.y;
			break;		
		case CRectTracker::hitLeft:
			rect.left += r.x;
			break;	
		case CRectTracker::hitMiddle:
		case 11:
			rect += r;
			break;
		}

		// Determine scaling in the x and y
		CPoint tr;
		double scaling_x = 1.0;
		double scaling_y = 1.0;

		scaling_x = fabs(rect.Width()) / fabs( a.x - b.x );
		scaling_y = fabs(rect.Height()) / fabs( a.y - b.y );

		a.x = rect.left;
		a.y = rect.top;
		b.x = rect.right;
		b.y = rect.bottom;

		// Now apply the scaling...
		arcpointCollection::iterator it = m_handles.begin();
		while (it != m_handles.end())
		{
			CArcPoint p = *it;
			p.x *= scaling_x;
			p.y *= scaling_y;
			*it = p;
			++ it;
		}

	}

	FlatternPath();
	Display();
}

int CDrawPolygon::SetCursorEdit( CPoint p )
{
	CRect r(a.x,a.y,b.x,b.y);
	r.NormalizeRect();

	int s = -1;
	
	s = IsInsideField(p);
	if (s >= 1000)
	{
		// Always display the same cursor
		s = 8;
	}

	if (s == -1 && IsInsidePolygon(p))
	{
		s = 11;
	}


	return s;
}


// Try and find out if the point is inside the polygon.
//
// We use the odd/even rule.  We draw an imaginary line
// in any direction and count the number of lines we
// cross, if it is odd then we are inside the polygon,
// if even then it is outside.
//
// We can choose the line direction, so we pick a
// horizontal line.  We determine the line crossing
// by converting the lines to a parametric version:
// x = gy + h
// l1.x - g * l1.y = h
// Then determine if x is lies within the line at the crossing
// point, by comparing it with the line ends.
//
BOOL CDrawPolygon::IsInsidePolygon( CPoint p )
{
	int s = 0;
	pointCollection::iterator it = m_points.begin();
	
	CDPoint l1,l2;
	int count = 0;

	// Pre-load with the last point, so
	// we check the full round polygon...
	l2 = m_points.back() + a;

	for (;it != m_points.end(); ++it, ++s)
	{
		// Get the line
		l1 = l2;
		l2 = *it + a;

		// We can ignore horizontal lines
		if (l1.y == l2.y)
		{
			continue;
		}

		// Is the line vertical?
		int x_intersect;
		if (l1.x == l2.x)
		{
			// Perform quick calculation
			if (   p.y < min(l1.y,l2.y) 
				|| p.y > max(l1.y,l2.y) )
			{
				continue;
			}
			x_intersect = l1.x;
		}
		else
		{
			// Now convert the line to a parametric
			// equation...
			double g = static_cast<double>(l2.x - l1.x) / static_cast<double>(l2.y - l1.y);
			double h = l2.x - g * l2.y;

			// Now find out where this intersects our
			// y co-ordinate
			// Note: we can have problems if the result lies
			// directly on a vertex.  To compensate we add 0.5 to the
			// test position.  As all vertex's y position are all ways
			// integer we can never lie on a vertex.
			x_intersect = static_cast<int>(g * (static_cast<double>(p.y) + 0.5) + h);
		}

		// Is this inside the line?
		if (	x_intersect < min(l1.x,l2.x) 
			||	x_intersect > max(l1.x,l2.x))
		{
			// No, so continue
			continue;
		}

		// Is this in the positive direction?
		if (x_intersect > p.x)
		{
			// Yes, so count it!
			count ++;
		}
	}

	return (count & 1) != 0;
}

void CDrawPolygon::AddPolyBezier( pointCollection &cp, CArcPoint p1, CArcPoint p2 )
{
 	double x, y;
  	double t, t2, t3, a, b, c;

	CDPoint Np[4];
	Np[0] = CDPoint(p1.x,p1.y);
	Np[3] = CDPoint(p2.x,p2.y);
	double mid_x = (Np[0].x + Np[3].x)/2.0;
	double mid_y = (Np[0].y + Np[3].y)/2.0;

	switch (p2.arc)
	{
	case CArcPoint::Arc_out:
		Np[1].x = Np[0].x;
		Np[1].y = mid_y;
		Np[2].x = mid_x;
		Np[2].y = Np[3].y;
		break;
	default:
		Np[1].x = mid_x;
		Np[1].y = Np[0].y;
		Np[2].x = Np[3].x;
		Np[2].y = mid_y;
		break;
	}

	cp.push_back(Np[0]);

  	for (t=0.0; t<=1.0; t += 0.05) 
	{
   		t2 = t * t;
   		t3 = t2 * t;
   		a = 1 - 3*t + 3*t2 - t3;
   		b = 3*(t - 2*t2 + t3);
   		c = 3*(t2 - t3);
   		x = a * Np[0].x
       		+ b * Np[1].x
       		+ c * Np[2].x
       		+ t3 * Np[3].x;
   		y = a * Np[0].y
       		+ b * Np[1].y
       		+ c * Np[2].y
       		+ t3 * Np[3].y;

		cp.push_back(CDPoint(x, y));
  	}


	cp.push_back(Np[3]);
}



void CDrawPolygon::FlatternPath()
{
	m_points.erase( m_points.begin(), m_points.end() );

	// Now inspect the handles and create a flat line
	// version
	arcpointCollection::iterator it = m_handles.begin();
	CArcPoint last;
	while (it != m_handles.end())
	{
		if ((*it).arc != CArcPoint::Arc_none)
		{
			AddPolyBezier( m_points, last, *it );
		}
		else
		{
			
			m_points.push_back( CDPoint( (*it).x, (*it).y ) );
		}

		(*it).flatterned_segment = m_points.size();

		last = *it;
		++ it;
	}
}


// Display the line on the screen!
void CDrawPolygon::Paint(Context &dc,paint_options options)
{
  dc.SelectPen(m_pDesign->GetOptions()->GetStyle(Style), options);
 
  dc.SetROP2(R2_COPYPEN);


  if (Fill == fsNONE || !m_segment || options == draw_selectable)
  {
	 dc.Polyline( m_points, a, NULL );

	 pointCollection cp;
	  if (!m_segment) 
	  {
		CArcPoint p1;
		if (m_points.size() > 0)
		{
			CDPoint bq = m_points.back();
			p1 = CArcPoint( bq.x, bq.y );
		}
			
		switch (g_EditToolBar.m_LineEdit.mode) 
		  {
			case 1: 
				cp.push_back( CDPoint( p1.x, p1.y ) );
				p1 = CArcPoint(b.x - a.x,p1.y);
				break;
			case 2:	
				cp.push_back( CDPoint( p1.x, p1.y ) );
				p1 = CArcPoint(p1.x,b.y - a.y);
				break;
		  }

		  CArcPoint p2(b.x - a.x, b.y - a.y,g_EditToolBar.m_LineEdit.GetArcType());
		  if (p2.arc != CArcPoint::Arc_none)
		  {
			  AddPolyBezier( cp, p1, p2 );
		  }
		  else
		  {
			  cp.push_back(CDPoint(p1.x,p1.y));
			  cp.push_back(CDPoint(p2.x,p2.y));
		  }


	  	 dc.Polyline( cp, a, NULL );

	  }

  }
  else
  {
	  dc.Polyline( m_points, a, m_pDesign->GetOptions()->GetFillStyle(Fill) );
  }
}


// Store the line in the drawing
CDrawingObject* CDrawPolygon::Store()
{
  CDrawPolygon *NewObject;

  NewObject = new CDrawPolygon(m_pDesign);
  // Copy the details
  m_segment=1;
  *NewObject = *this;

  m_pDesign->Add(NewObject);

  return NewObject;
}


void CDrawPolygon::BeginEdit(BOOL re_edit)
{
  m_segment=1;

  m_pDesign->GetOptions()->SetCurrentStyle(Style);
  m_pDesign->GetOptions()->SetCurrentFillStyle(Fill);


  m_re_edit = re_edit;
  if (re_edit)
  {
	  g_EditToolBar.m_PolygonEdit.Open(m_pDesign,this);
	  g_EditToolBar.m_PolygonEdit.mode=0;
  }
  else
  {
	  g_EditToolBar.m_LineEdit.Open(m_pDesign,this);
	  g_EditToolBar.m_LineEdit.mode=0;
  }
}

void CDrawPolygon::EndEdit()
{
  Display();
  if (m_re_edit)
  {
	g_EditToolBar.m_PolygonEdit.Close();
  }
  else
  {
	  g_EditToolBar.m_LineEdit.Close();
  }

  m_re_edit = FALSE;
}


int CDrawPolygon::IsInsideLine(int left,int right,int top,int bottom)
{
	int s = 0;
	CPoint la;
	pointCollection::iterator it = m_points.begin();
	while (it != m_points.end())
	{
		CDPoint a_np = *it + a;
		CPoint np(a_np.x, a_np.y);
		if (s != 0)
		{
			CLineUtils l( la, np);
			BOOL r = l.IsInside(left,right,top,bottom);
			if (r)
			{
			  return TRUE;
			}
		}

		la = np;
		++ s;
		++ it;
	}

	return FALSE;
}


double CDrawPolygon::DistanceFromPoint( CPoint p )
{
	// Are we filled?
	if (Fill == fsNONE)
	{
		// Ok, so check for distance from one of our lines...
		double closest_distance = 100.0;
		int s = 0;
		CPoint la;
		pointCollection::iterator it = m_points.begin();
		while (it != m_points.end())
		{
			CDPoint a_np = *it + a;
			CPoint np(a_np.x, a_np.y);
			if (s != 0)
			{
				CLineUtils l( la, np);
				CPoint d;
				closest_distance = min( closest_distance, l.DistanceFromPoint( p, d ) );
			}

			la = np;
			++ s;
			++ it;
		}

		return closest_distance;
	}
	else
	{
		// There is a fill, so just use the normal isinside routine...
		return IsInside( p.x,p.x,p.y,p.y ) ? 0 : 100.0;
	}

	return 100.0;
}



BOOL CDrawPolygon::IsInside(int left,int right,int top,int bottom)
{
	// We are definately inside if we intersect a line...
	if (IsInsideLine( left,right,top,bottom ))
	{
		return TRUE;
	}

	// If we are filled or editing then check to see 
	// if any of the four points are inside our polygon
	if (Fill != fsNONE || m_re_edit)
	{
		return IsInsidePolygon( CPoint(left,top) )
		|| IsInsidePolygon( CPoint(right,top) )
		|| IsInsidePolygon( CPoint(right,bottom) )
		|| IsInsidePolygon( CPoint(left,bottom) );
	}

	return FALSE;
}



