/*
	TinyCAD program for schematic capture
	Copyright 1994/1995/2002,2003 Matt Pyne.

	This program is free software; you can redistribute it and/or
	modify it under the terms of the GNU Lesser General Public
	License as published by the Free Software Foundation; either
	version 2.1 of the License, or (at your option) any later version.

	This program is distributed in the hope that it will be useful,
	but WITHOUT ANY WARRANTY; without even the implied warranty of
	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
	Lesser General Public License for more details.

	You should have received a copy of the GNU Lesser General Public
	License along with this library; if not, write to the Free Software
	Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
*/

#include "stdafx.h"
#include <winreg.h>
#include "registry.h"

//=====================================================================
//== init static variables                                           ==
//=====================================================================
const CString	Registry::M_SKEY	= "Software\\TinyCAD\\TinyCAD\\1x20";
CRegKey			Registry::m_oKey	= CRegKey();
Registry		Registry::m_oDummi;

//-------------------------------------------------------------------------
//-- Attach this object to the registry
Registry::Registry()
{
	m_oKey.Create( HKEY_CURRENT_USER, M_SKEY );
	
	// Does the registry information exist - if not create it
	if( ! keyExists("PageSize") )
	{
		CreateDefaultEntries();
	}
}
//-------------------------------------------------------------------------
//-- Write the initial data to the registry
void Registry::CreateDefaultEntries()
{
	// Save the page size setup
	Set( "PageSize", CSize(297 * PIXELSPERMM, 210 * PIXELSPERMM) );
	Set( "Maximize", false );

  	TCHAR theBuffer[1024];
	DWORD theBytes = GetModuleFileName( NULL,theBuffer,sizeof(theBuffer) - 1 );

	if( theBytes != 0 )
	{
		TCHAR* thePtr = theBuffer + theBytes;

		while (thePtr >= theBuffer && *thePtr != _TCHAR('\\'))
		{
			thePtr--;
		}
		*thePtr = _TCHAR('\0');
	
		CString theMainDir = CString( theBuffer ) + "\\";

		// Search the default library directory for libraries
		for (int l=0;l<2;l++)
		{
			CString sSearch = theMainDir + "library\\";
		
			if( l == 0 )
			{
				sSearch += "*.idx";
			}
			else
			{
				sSearch += "*.mdb";
			}

			FindFile theFind( sSearch );

			if (theFind.Success()) 
			{
				CString Libraries = "";

				do {
					// Add this library to the list of libraries in use
					if (Libraries != "")
						Libraries += ",";
					Libraries += theMainDir+"library\\"+theFind.GetName();;
					// Remove the extension (of .idx)
					Libraries = Libraries.Left(Libraries.ReverseFind('.'));
				} while (theFind.FindNext());

				Set( "Libraries", Libraries );
			}
		}

		// Finially place the association of this file and the design type
		Associate( ".dsn","TinyCAD design", "TinyCAD" );
	}
}
//-------------------------------------------------------------------------
//-- Close the registry database                                         --
//-------------------------------------------------------------------------
Registry::~Registry()
{
	m_oKey.Close();
}
//-------------------------------------------------------------------------

//-------------------------------------------------------------------------
//-- Associate this software with this key
void Registry::Associate
(
	CString	sExtension,
	CString	sTitle,
	CString	sProgram
)
{
	// a separate key
	CRegKey oRegKey;

	// Associate the file type with an extension
	oRegKey.Create( HKEY_CLASSES_ROOT, sExtension );
	oRegKey.SetValue( HKEY_CLASSES_ROOT, sExtension, sTitle );

	// Write the description for this key
	oRegKey.SetValue( HKEY_CLASSES_ROOT, sTitle, sTitle );

	// Associate the file type with a program
	oRegKey.Create( HKEY_CLASSES_ROOT, sTitle + "\\shell\\open\\command" );
	oRegKey.SetValue( HKEY_CLASSES_ROOT, sTitle + "\\shell\\open\\command", sProgram + " %1" );
		
	// Associate the file with a program for printing
	oRegKey.Create( HKEY_CLASSES_ROOT, sTitle + "\\shell\\print\\command" );
	oRegKey.SetValue( HKEY_CLASSES_ROOT, sTitle + "\\shell\\print\\command", sProgram + " /p %1" );
		
	oRegKey.Close();
}
//-------------------------------------------------------------------------
int Registry::QueryValueEx(CString sKey,DWORD* Type,unsigned char *data,unsigned long *size)
{
	return RegQueryValueEx( (HKEY) m_oKey, sKey, NULL, Type, (unsigned char *)data, size);
}
//-------------------------------------------------------------------------
int Registry::SetValueEx(CString sKey, DWORD Type, unsigned char *data, unsigned long length)
{
	return RegSetValueEx( (HKEY) m_oKey, sKey, 0, Type, data, length);
}
//-------------------------------------------------------------------------
// Place a CString item
void Registry::Set( CString sKey, CString sVal )
{
	m_oKey.SetValue( sVal, sKey );
}
//-------------------------------------------------------------------------
// Place a DWORD item
void Registry::Set( CString sKey, DWORD nVal )
{
	m_oKey.SetValue( nVal, sKey );
}
//-------------------------------------------------------------------------
// Place a WORD item
void Registry::Set( CString sKey, WORD nVal )
{
	m_oKey.SetValue( nVal, sKey );
}
//-------------------------------------------------------------------------
// Place int and BOOL items
void Registry::Set( CString sKey, int nVal )
{
	m_oKey.SetValue( nVal, sKey );
}
//-------------------------------------------------------------------------
// Place a bool value
void Registry::Set( CString sKey, bool bVal )
{
	DWORD nVal = bVal ? 1 : 0;

	Set( sKey, nVal );
}
//-------------------------------------------------------------------------
// Place a raw data block item
void Registry::Set( CString sKey, void* data,int length )
{
	SetValueEx(sKey, REG_BINARY, (unsigned char *)((char *)data), length);
}
//-------------------------------------------------------------------------
// Place a CSize 
void Registry::Set( CString sKey, CSize szVal )					
{
	CString sPageSize = "";

	sPageSize.Format( "%d,%d", szVal.cx, szVal.cy );  

	Set( "PageSize", sPageSize );
}
//-------------------------------------------------------------------------
// Get a sub item from the registry
bool Registry::GetRaw( CString sKey,char *data,unsigned int length)
{
  DWORD type;
  DWORD size	=length;
  int	r		= QueryValueEx( sKey, &type,(unsigned char *)data, &size);

  return r == ERROR_SUCCESS;
}
//-------------------------------------------------------------------------
// Get a String subkey 
CString	Registry::GetString( CString sKey, CString sDefaultVal )
{
	CString sReturn = sDefaultVal;
	DWORD	nSize	= 1024;
	char*	pTest	= (char*) sReturn.GetBuffer( nSize );
	
	m_oKey.QueryValue( pTest, sKey, &nSize );

	sReturn.ReleaseBuffer();

	return sReturn;
}
//-------------------------------------------------------------------------
// Get a String array
// Splits a comma separated string into its substrings
CStringList* Registry::GetStringList( CString sKey )
{
	CStringList* colReturn	= new CStringList();

	// Is there a list of items in the registry?
	CString		sItems		= Registry::GetString( sKey, "" );

	// single item
	CString		sItem		= "";

	// Comma position
	int			nFoundSep	= -1;


	if( sItems != "" ) 
	{
		while( sItems.GetLength() !=0 ) 
		{
			nFoundSep = sItems.Find( ',' );

			if( nFoundSep==-1 ) 
			{
				sItem	= sItems;
				sItems	= "";
			} 
			else 
			{
				sItem	= sItems.Left( nFoundSep );
				sItems	= sItems.Mid ( nFoundSep + 1 );
			}

			colReturn->AddTail( sItem );
		}
	}

	return colReturn;
}
//-------------------------------------------------------------------------
bool Registry::GetBool( CString sKey, bool bDefaultVal )
{
	bool	bReturn = bDefaultVal;
	DWORD	nVal	= 0;
	
	return GetInt( sKey, bDefaultVal ) != 0; 
}
//-------------------------------------------------------------------------
int Registry::GetInt( CString sKey, int nDefaultVal )
{
	int		nReturn = nDefaultVal;
	DWORD	nVal	= 0;
	

	/* We can't use this because it ASSERTs if the type is not DWORD 
	if( m_oKey.QueryValue( nVal, sKey ) == ERROR_SUCCESS )
	{
		nReturn = (int) nVal;
	}
	*/

	DWORD dwType = NULL;
	DWORD dwCount = sizeof(nVal);
	LONG lRes = RegQueryValueEx(m_oKey.m_hKey, sKey, NULL, &dwType,
		(LPBYTE)&nVal, &dwCount);

	if (lRes == ERROR_SUCCESS)
	{
		return nVal;
	}
	
	return nReturn;
}
//-------------------------------------------------------------------------
BOOL Registry::GetBOOL( CString sKey, BOOL nDefaultVal )
{
	BOOL	nReturn = nDefaultVal;
	DWORD	nVal	= 0;
	
	if( m_oKey.QueryValue( nVal, sKey ) == ERROR_SUCCESS )
	{
		nReturn = (BOOL) nVal;
	}
	
	return nReturn;
}
//-------------------------------------------------------------------------
char Registry::GetChar( CString sKey, char nDefaultVal )
{
	char		nReturn = nDefaultVal;
	DWORD		nVal	= 0;
	
	if( m_oKey.QueryValue( nVal, sKey ) == ERROR_SUCCESS )
	{
		nReturn = (char) nVal;
	}
	
	return nReturn;
}
//-------------------------------------------------------------------------
LONG Registry::GetLong( CString sKey, LONG nDefaultVal )
{
	LONG		nReturn = nDefaultVal;
	DWORD		nVal	= 0;
	
	if( m_oKey.QueryValue( nVal, sKey ) == ERROR_SUCCESS )
	{
		nReturn = (LONG) nVal;
	}
	
	return nReturn;
}
//-------------------------------------------------------------------------
WORD Registry::GetWord( CString sKey, WORD nDefaultVal )
{
	WORD	nReturn = nDefaultVal;
	DWORD	nVal	= 0;
	
	if( m_oKey.QueryValue( nVal, sKey ) == ERROR_SUCCESS )
	{
		nReturn = (WORD) nVal;
	}
	
	return nReturn;
}
//-------------------------------------------------------------------------
CSize Registry::GetSize( CString sKey, CSize szDefault )
{
	CSize	szReturn;
	CString sSize		= GetString( sKey, "" );
	int		nFoundComma	= -1;

	if( sSize != "" ) 
	{
		nFoundComma = sSize.Find( ',' );
		
		if( nFoundComma != -1 )
		{
			szReturn.cy = atoi( sSize.Mid(nFoundComma+1) );
			szReturn.cx = atoi( sSize.Left(nFoundComma) );
		}

	}
	
	return szReturn;
}
//-------------------------------------------------------------------------
bool Registry::keyExists( CString sKey )
{
	DWORD	nSize	= 1024;
	CString	sDummi	= "";
	char*	pDummi	= (char*) sDummi.GetBuffer( nSize );
	bool	bReturn = m_oKey.QueryValue( pDummi, sKey, &nSize ) == ERROR_SUCCESS;

	sDummi.ReleaseBuffer();
	
	return bReturn;
}
//-------------------------------------------------------------------------
//-- These functions use the registry 
void Registry::SetPrinterMode( HANDLE DevMode, HANDLE PrinterName )
{
	DEVMODE*	theDevMode		= (DEVMODE *)::GlobalLock(DevMode);
	DEVNAMES*	thePrinterName	= (DEVNAMES *)::GlobalLock(PrinterName);

  	if( theDevMode != NULL )
	{
		Set("PrinterMode",(void *)theDevMode,theDevMode->dmSize+theDevMode->dmDriverExtra);
	}

	if( thePrinterName != NULL ) 
	{
		CString theNames;

		theNames = 	((char *)thePrinterName)+thePrinterName->wDriverOffset;
		theNames += ",";
		theNames += ((char *)thePrinterName)+thePrinterName->wDeviceOffset;
		theNames += ",";
		theNames += ((char *)thePrinterName)+thePrinterName->wOutputOffset;
		
		Set( "Printer", theNames );
	}

	::GlobalUnlock(DevMode);
	::GlobalUnlock(PrinterName);
}
//-------------------------------------------------------------------------



////// The find files class //////


// Find the first file

FindFile::FindFile(CString FileSpec)
{
  handle = ::FindFirstFile(FileSpec, &fileinfo );
}


// Find the next file
bool FindFile::FindNext()
{
  return (::FindNextFile(handle,&fileinfo) != 0 );
}

// Close the FindFile structure
FindFile::~FindFile()
{
	::FindClose(handle);	
}


char *FindFile::GetName()
{
	return fileinfo.cFileName;	
}


// Was the find file successfull?
bool FindFile::Success()
{
	return handle != INVALID_HANDLE_VALUE;
}
