/*
	TinyCAD program for schematic capture
	Copyright 1994/1995/2002,2003 Matt Pyne.

	This program is free software; you can redistribute it and/or
	modify it under the terms of the GNU Lesser General Public
	License as published by the Free Software Foundation; either
	version 2.1 of the License, or (at your option) any later version.

	This program is distributed in the hope that it will be useful,
	but WITHOUT ANY WARRANTY; without even the implied warranty of
	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
	Lesser General Public License for more details.

	You should have received a copy of the GNU Lesser General Public
	License along with this library; if not, write to the Free Software
	Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
*/


#include "stdafx.h"
#include "TinyCad.h"
#include "TinyCadView.h"
#include "resource.h"
#include "revision.h"
#include "colour.h"
#include "option.h"
#include "registry.h"
#include "layer.h"
#include "LibraryCollection.h"
#include <list>


////// The resource methods //////

// The constructor
CDocResource::CDocResource() 
{ 
	ResourceNumber = 0; 
	MergeNumber = 0; 
	InUse = FALSE; 
	next = NULL; 
}

// The virtual destructor
CDocResource::~CDocResource() 
{ 
	// Delete down the chain...
	delete next;
}


hRESOURCE CDocResource::ConvertOld(hRESOURCE n)
{
	CDocResource *pointer = this;

	while (pointer!=NULL && n!=pointer->MergeNumber) {
		pointer = pointer->next;
	}

	// If the number was out of range then return the default font
	if (pointer == NULL)
		return 0;

	return pointer->ResourceNumber;
}


// Add a new resource to this list
hRESOURCE CDocResource::Add(CDocResource *NewResource,hRESOURCE NewMerge)
{
	CDocResource *pointer = this, *last = NULL;
	hRESOURCE nextNumber = 0;

	while (pointer!=NULL && !(pointer->MergeNumber == -1 && pointer->Compare(NewResource)) ) 
	{
		// Find the next index number available
		if (pointer->ResourceNumber >= nextNumber)
			nextNumber = pointer->ResourceNumber + 1;

		last = pointer;
		pointer=pointer->next;
	}

	// This resouce is already defined
	if (pointer != NULL) {
		pointer->MergeNumber = NewMerge;
		// The new resource is not required and can be deleted!
		delete NewResource;
		// Now return the resource number
		return pointer->ResourceNumber;
	}

	ASSERT(last != NULL);

	// Not defined so add the resource to the end of the list
	NewResource->ResourceNumber = nextNumber;
	NewResource->MergeNumber = NewMerge;
	last->next = NewResource;

	return nextNumber;
}


// Get a resource from the list
CDocResource *CDocResource::Get(hRESOURCE n)
{
	CDocResource *pointer = this;

	while (pointer != NULL) {
		if (pointer->ResourceNumber == n)
			break;
		pointer = pointer->next;
	}

	// If the number was out of range then return the default font (the head of the list)
	if (pointer == NULL)
		pointer = this;

	return pointer;

}

// Tag a resource given a resource number
void CDocResource::Tag(hRESOURCE n)
{
	CDocResource *pointer = this;

	while (pointer != NULL) {
		if (pointer->ResourceNumber == n)
			break;
		pointer = pointer->next;
	}

	// If the number was out of range then return the default font
	if (pointer != NULL)
		pointer->InUse = TRUE;
}


void CDocResource::UnTag()
{
	CDocResource *pointer = this;

	while (pointer != NULL) {
		pointer->InUse = FALSE;
		pointer = pointer->next;
	}

}



void CDocResource::SaveXML( CTinyCadDoc *pDesign, const char *name, CXMLWriter &xml)
{
	// Count the number of fonts in operation
	hRESOURCE NumberOfResources = 0;
	CDocResource  *pointer = this;

	std::list<CDocResource *> resources;
	while (pointer!=NULL) 
	{
		if (pointer->InUse)
		{
			resources.push_back( pointer );
			NumberOfResources++;
		}
		pointer = pointer->next;
	}

	// Now actually save the resources
	std::list<CDocResource *>::iterator it = resources.begin();

	while (it != resources.end()) 
	{
		pointer = *it;

		xml.addTag( name );
		xml.addAttribute( "id", pointer->ResourceNumber );
		pointer->SaveItemXML(pDesign,xml);
		xml.closeTag();

		it ++;
	}
}

// Clear the merge numbers
void CDocResource::ResetMerge()
{
	// All the merge numbers to invalid
	CDocResource *pointer = this;
	while (pointer!=NULL) 
	{
		pointer->MergeNumber = -1;
		pointer = pointer->next;
	}
}


void CDocResource::Load(CTinyCadDoc *pDesign, CStream &theFile)
{
	// All the merge numbers to invalid
	ResetMerge();

	// Read in the number of Fonts in this list
	WORD NumberOfResources;
	theFile >> NumberOfResources;

	hRESOURCE OldResourceNumber = 0;

	while (NumberOfResources > 0) 
	{
		theFile >> OldResourceNumber;
		LoadItem(pDesign, theFile,OldResourceNumber);
		NumberOfResources--;
	}
}






////// The Line Style Options /////


// The resource for line styles
BOOL ListOfStyles::Compare(CDocResource *o)
{
	#define  pStyle (((ListOfStyles *)o)->theLineStyle)

	return (pStyle.Style     == theLineStyle.Style
		 && pStyle.Thickness == theLineStyle.Thickness
		 && pStyle.Colour    == theLineStyle.Colour );
}


void ListOfStyles::SaveItemXML( CTinyCadDoc *pDesign, CXMLWriter &xml)
{
	xml.addTag( "STYLE", theLineStyle.Style );
	xml.addTag( "COLOR", theLineStyle.Colour );
	xml.addTag( "THICKNESS", theLineStyle.Thickness );
}

void ListOfStyles::LoadItemXML(CTinyCadDoc *pDesign, CXMLReader &xml)
{
	hRESOURCE n;
	CString name;
	LineStyle ls;
	memset( &ls, 0, sizeof( LineStyle ) );

	xml.getAttribute( "id", n );
	xml.intoTag();
	
	while (xml.nextTag( name ))
	{
		if (name == "STYLE")
		{
			xml.getChildData(ls.Style);
		}
		else if (name == "COLOR")
		{
			xml.getChildData(ls.Colour);
		}
		else if (name == "THICKNESS")
		{
			xml.getChildData(ls.Thickness);
		}
	}

	xml.outofTag();

	ListOfStyles *nFS = new ListOfStyles(ls);
	Add(nFS,n);
}



void ListOfStyles::LoadItem(CTinyCadDoc *pDesign, CStream &theFile, hRESOURCE n)
{
	LineStyle ls;
	theFile.Read(&ls,sizeof(LineStyle));

	ListOfStyles *nFS = new ListOfStyles(ls);

	Add(nFS,n);
}




////// The Line Style Options /////


// The resource for fill styles
BOOL ListOfFillStyles::Compare(CDocResource *o)
{
	#define  pFillStyle (((ListOfFillStyles *)o)->theFillStyle)

	return pFillStyle.Colour    == theFillStyle.Colour
		&& pFillStyle.Index		== theFillStyle.Index;
}


void ListOfFillStyles::SaveItemXML( CTinyCadDoc *pDesign, CXMLWriter &xml)
{
	xml.addTag( "INDEX", theFillStyle.Index );
	xml.addTag( "COLOR", theFillStyle.Colour );
}

void ListOfFillStyles::LoadItemXML(CTinyCadDoc *pDesign, CXMLReader &xml)
{
	hRESOURCE n;
	FillStyle ls;
	CString name;

	memset( &ls, 0, sizeof( FillStyle ) );

	xml.getAttribute( "id", n );
	xml.intoTag();
	
	while (xml.nextTag( name ))
	{
		if (name == "INDEX")
		{
			xml.getChildData(ls.Index);
		}
		else if (name == "COLOR")
		{
			xml.getChildData(ls.Colour);
		}
	}

	xml.outofTag();

	ListOfFillStyles *nFS = new ListOfFillStyles(ls);
	Add(nFS,n);
}



void ListOfFillStyles::LoadItem(CTinyCadDoc *pDesign, CStream &theFile, hRESOURCE n)
{
	FillStyle ls;
	theFile.Read(&ls,sizeof(theFillStyle));

	ListOfFillStyles *nFS = new ListOfFillStyles(ls);

	Add(nFS,n);
}




////// The MetaFile Resource /////


// The resource for line styles
BOOL ListOfMetaFiles::Compare(CDocResource *o)
{
	#define  pMetaFile (((ListOfMetaFiles *)o)->theMetaFile)

	return theMetaFile == pMetaFile;
}


void ListOfMetaFiles::SaveItemXML( CTinyCadDoc *pDesign, CXMLWriter &xml)
{
  // Put the meta-file into a memory buffer
  UINT size = GetEnhMetaFileBits( theMetaFile, 0, NULL );
  BYTE *buffer = new BYTE[ size ];
  GetEnhMetaFileBits( theMetaFile, size, buffer );

  // Write the size to the archive
  xml.addChildDataUUencode( buffer, size );

  // Now discard the buffer
  delete[] buffer;
}


void ListOfMetaFiles::LoadItemXML(CTinyCadDoc *pDesign, CXMLReader &xml)
{
	hRESOURCE n;
	CString name;

	xml.getAttribute( "id", n );

	UINT size = 0;
	BYTE *buffer = NULL;

	// Get the metafile from the buffer
	xml.intoTag();
	
	while (xml.nextTag( name ))
	{
		if (name == "UUENCODE")
		{
			xml.getChildDataUUdecode( buffer, size );
		}
	}

	xml.outofTag();

	// Now create the new metafile...
	HENHMETAFILE data = SetEnhMetaFileBits( size, buffer );

	// ... and delete the buffer
	delete[] buffer;

	ListOfMetaFiles *nFS = new ListOfMetaFiles(data);

	Add(nFS,n);
}

void ListOfMetaFiles::LoadItem(CTinyCadDoc *pDesign, CStream &theFile, hRESOURCE n)
{
  // Now read in the metafile...
  UINT size;
  theFile >> size;
  BYTE *buffer = new BYTE[ size ];

  // Get the metafile from the buffer
  theFile.Read( buffer, size );

  // Now create the new metafile...
  HENHMETAFILE data = SetEnhMetaFileBits( size, buffer );

  // ... and delete the buffer
  delete[] buffer;

  ListOfMetaFiles *nFS = new ListOfMetaFiles(data);

  Add(nFS,n);
}





////// The Font choosing Options //////


ListOfFonts::ListOfFonts(LOGFONT f)
{ 
	FontStyle = f; 
}


// The virtual constructs for the ListOfFonts resource
BOOL ListOfFonts::Compare(CDocResource *o)
{
	#define  pFont ((ListOfFonts *)o)->FontStyle

	return pFont.lfHeight    == FontStyle.lfHeight
		&& pFont.lfWidth     == FontStyle.lfWidth
  		&& pFont.lfWeight    == FontStyle.lfWeight
  		&& pFont.lfItalic    == FontStyle.lfItalic
  		&& pFont.lfUnderline == FontStyle.lfUnderline
  		&& pFont.lfStrikeOut == FontStyle.lfStrikeOut
  		&& pFont.lfCharSet   == FontStyle.lfCharSet
  		&& strcmp(pFont.lfFaceName,FontStyle.lfFaceName)==0;
}


void ListOfFonts::SaveItemXML( CTinyCadDoc *pDesign, CXMLWriter &xml)
{
	xml.addTag( "HEIGHT", FontStyle.lfHeight);
	xml.addTag( "WIDTH", FontStyle.lfWidth);
  	xml.addTag( "WEIGHT", FontStyle.lfWeight);
  	xml.addTag( "ITALIC", FontStyle.lfItalic);
  	xml.addTag( "UNDERLINE", FontStyle.lfUnderline);
  	xml.addTag( "STRIKEOUT", FontStyle.lfStrikeOut);
  	xml.addTag( "CHARSET", FontStyle.lfCharSet);
  	xml.addTag( "FACENAME", FontStyle.lfFaceName);
}

void ListOfFonts::LoadItemXML(CTinyCadDoc *pDesign, CXMLReader &xml)
{
	hRESOURCE n;
	LOGFONT lf;
	CString name;

	memset( &lf, 0, sizeof( LOGFONT ) );

	xml.getAttribute( "id", n );
	xml.intoTag();
	
	while (xml.nextTag( name ))
	{
		if (name == "HEIGHT")
		{
			xml.getChildData(lf.lfHeight);
		} 
		else if (name == "WIDTH")
		{
			xml.getChildData(lf.lfWidth);
		}
		else if (name == "WEIGHT")
		{
			xml.getChildData(lf.lfWeight);
		}
		else if (name == "ITALIC")
		{
			xml.getChildData(lf.lfItalic);
		}
		else if (name == "UNDERLINE")
		{
			xml.getChildData(lf.lfUnderline);
		}
		else if (name == "STRIKEOUT")
		{
			xml.getChildData(lf.lfStrikeOut);
		}
		else if (name == "CHARSET")
		{
			xml.getChildData(lf.lfCharSet);
		}
		else if (name == "FACENAME")
		{
			CString s;
			xml.getChildData(s);
			strcpy(lf.lfFaceName, s );
		}
	}

	xml.outofTag();

	ListOfFonts *nFS = new ListOfFonts(lf);
	Add(nFS,n);
}



void ListOfFonts::LoadItem(CTinyCadDoc *pDesign, CStream &theFile, hRESOURCE n)
{
	LOGFONT lf;

	theFile.Read(&lf,sizeof(LOGFONT));

	ListOfFonts *nFS = new ListOfFonts(lf);

	Add(nFS,n);
}






////// The Symbol Resource /////

ListOfSymbols::~ListOfSymbols()
{
	delete theSymbol;
}

// The resource for line styles
BOOL ListOfSymbols::Compare(CDocResource *o)
{
	CDesignFileSymbol *pSymbol = static_cast<CDesignFileSymbol*>(((ListOfSymbols *)o)->theSymbol);
	return theSymbol->name == pSymbol->name && theSymbol->FilePos == pSymbol->FilePos;
}


void ListOfSymbols::SaveItemXML( CTinyCadDoc *pDesign, CXMLWriter &xml)
{
	// Save the symbol with the document...
	theSymbol->SaveXML( xml );
}

void ListOfSymbols::LoadItemXML(CTinyCadDoc *pDesign, CXMLReader &xml)
{
	hRESOURCE n;
	CString name;
	drawingCollection method;
	CDesignFileSymbol *psymbol = new CDesignFileSymbol( pDesign, method, CPoint(0,0) );

	xml.getAttribute( "id", n );
	xml.intoTag();
	
	psymbol->LoadXML( pDesign, xml );
	
	xml.outofTag();

	ListOfSymbols *nFS = new ListOfSymbols(psymbol);
	Add(nFS,n);
}



void ListOfSymbols::LoadItem(CTinyCadDoc *pDesign, CStream &theFile, hRESOURCE n)
{
	// Save the symbol with the document...
	drawingCollection method;
	CDesignFileSymbol *psymbol = new CDesignFileSymbol( pDesign, method, CPoint(0,0) );
	psymbol->Load( pDesign, theFile );

  ListOfSymbols *nFS = new ListOfSymbols(psymbol);

  Add(nFS,n);

}

// Reload from the library
void ListOfSymbols::ResetSymbol()
{
	// Extract this symbol from the library and see what happens...
	CLibraryStoreSymbol *new_symbol = CLibraryCollection::GetSymbol( theSymbol->name );


	if (new_symbol)
	{
		CTinyCadDoc *pDesign = theSymbol->GetDesign();
		delete theSymbol;
		theSymbol = new_symbol->GetDesignSymbol( pDesign );
	}
}
