/*
	TinyCAD program for schematic capture
	Copyright 1994-2004 Matt Pyne.

	This program is free software; you can redistribute it and/or
	modify it under the terms of the GNU Lesser General Public
	License as published by the Free Software Foundation; either
	version 2.1 of the License, or (at your option) any later version.

	This program is distributed in the hope that it will be useful,
	but WITHOUT ANY WARRANTY; without even the implied warranty of
	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
	Lesser General Public License for more details.

	You should have received a copy of the GNU Lesser General Public
	License along with this library; if not, write to the Free Software
	Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
*/

// XMLReader.cpp: implementation of the CXMLReader class.
//
//////////////////////////////////////////////////////////////////////

#include "stdafx.h"
#include "tinycad.h"
#include "XMLReader.h"
#include "XMLException.h"

#ifdef _DEBUG
#undef THIS_FILE
static char THIS_FILE[]=__FILE__;
#define new DEBUG_NEW
#endif

//////////////////////////////////////////////////////////////////////
// Construction/Destruction
//////////////////////////////////////////////////////////////////////

CXMLReader::CXMLReader(CStream* pInput)
{
	m_pInput = pInput;
	m_uu_data = NULL;
	m_uu_size = 0;
	m_uu_state = 0;

	intoTag();
}

CXMLReader::~CXMLReader()
{
	while (m_tags.size() > 0)
	{
		delete m_tags.front();
		m_tags.pop_front();
	}
}


xml_parse_tag* CXMLReader::get_current_tag()
{
	return m_tags.front();
}


// Get the attribute information associated with the current
// tag...
bool CXMLReader::internal_getAttribute( const char *name, CString &data )
{
	xml_parse_tag::attrCollection::iterator it = get_current_tag()->m_attributes.find( name );
	if (it != get_current_tag()->m_attributes.end())
	{
		data = (*it).second;
		return true;
	}

	return false;
}


void CXMLReader::child_data( const char *in )
{
	if (m_uu_data != NULL && m_uu_size > 0)
	{
		int l = strlen( in );
		while (l > 0)
		{
			uudecode( in[0] );
			in ++;
			l --;
		}
	}
	else
	{
		m_child_data += in;
	}
}


// Scan and find the next tag
bool CXMLReader::getNextTag( CString &name )
{
	// Is the sub-buffer for the child data
	const int build_len = 255;
	char	child_data_build[ build_len + 1 ];
	int		child_data_index = 0;

	// Now scan to find the next tag...
	char c;
	for (;;)
	{
		do
		{
			if (m_pInput->Read(&c,1) != 1)
			{
				name = "";
				return false;
			}

			if (c != '<')
			{
				if (child_data_index == build_len || c == '&' )
				{
					child_data_build[ child_data_index ] = 0;
					child_data( child_data_build );
					child_data_index = 0;
				}

				if (c == '&')
				{
					child_data( xml_parse_tag::read_entity( m_pInput ) );
				}
				else
				{
					child_data_build[ child_data_index ] = c;
					child_data_index ++;
				}
			}
		}
		while (c != '<');

		// Now we have found the opening for the next
		// tag, we must scan it...
		get_current_tag()->parse( m_pInput );

		// Was this a comment?
		if (get_current_tag()->m_comment)
		{
			continue;
		}

		// Was this a closing tag?
		if (get_current_tag()->m_closing_tag)
		{
			// Yep, so this is the end of this run....
			name = get_current_tag()->m_tag_name;

			child_data_build[ child_data_index ] = 0;
			child_data( child_data_build );

			return false;
		}
		else
		{
			// Ok, we have a new opening tag...
			name = get_current_tag()->m_tag_name;

			child_data_build[ child_data_index ] = 0;
			child_data( child_data_build );

			return true;
		}
	}
}


// Scan until the end of the current tag...
//
bool CXMLReader::closeTag()
{
	// Is this tag self-closing?
	if (get_current_tag()->m_self_closing_tag
		|| get_current_tag()->m_closing_tag)
	{
		return true;
	}


	CString close_name = get_current_tag()->m_tag_name;

	// Nope, we must scan until we find the closing for
	// this tag ( recursively going into tags and out
	// again until we find our match...)
	for (;;)
	{
		CString name;

		// Is this the closing tag for the current tag?
		if (!getNextTag(name))
		{
			if (close_name == name)
			{
				return true;
			}
			else
			{
				throw new CXMLException(ERR_XML_WRONG_CLOSE, TRUE);
			}
		}
		else
		{
			// Must be an opening tag, so we need to close this
			// tag now...
			intoTag();
			outofTag();
		}
	}

	return true;
}


// Find the next peer tag of this tag
// returns false if there are no more tags at this
// level...
bool CXMLReader::nextTag( CString &name )
{
	// First, we must find the closing of the 
	// previous tag
	if (!closeTag())
	{
		return false;
	}

	m_child_data = "";
	return getNextTag( name );
}

// Move inside the current tag.  nextTag will then return
// tags that are inside the current tag....
void CXMLReader::intoTag()
{
	// Create a new parser to go into the tag system...
	xml_parse_tag *tag = new xml_parse_tag();
	m_tags.push_front( tag );
}




// Move up one level.  nextTag will then return tags
// that are one above the current level...
//
void CXMLReader::outofTag()
{

	// Have we already scanned to the closing tag?
	if (get_current_tag()->m_closing_tag && m_tags.size() > 1)
	{
		CString check_name = get_current_tag()->m_tag_name;

		// Copy this tag up one level...
		xml_parse_tag *ptag = m_tags.front();
		m_tags.pop_front();
		
		delete m_tags.front();
		m_tags.pop_front();

		m_tags.push_front( ptag );

		if (get_current_tag()->m_tag_name != check_name)
		{
			throw new CXMLException(ERR_XML_WRONG_CLOSE, TRUE);
		}
	}
	else
	{
		delete m_tags.front();
		m_tags.pop_front();

		closeTag();
	}

}


// Get the next child data associated with the current tag
// 
CString CXMLReader::internal_getChildData()
{
	m_child_data = "";

	closeTag();

	return m_child_data;
}

void CXMLReader::getChildDataUUdecode( BYTE* &data, UINT &size )
{
	// Get the size of this data
	CString name;

	getAttribute( "size", size );
	data = new BYTE[ size ];

	m_uu_size = size;
	m_uu_data = data;

	closeTag();
	
	m_uu_data = NULL;
	m_uu_size = 0;
	m_uu_state = 0;
}

// single character decode 
#define DEC(c)	(((c) - ' ') & 077)

// UUdecode
void CXMLReader::uudecode( char in )
{
	switch (m_uu_state)
	{
	case 0:	// We are awaiting the line count
		if (xml_parse_tag::is_whitespace(in))
		{
			break;
		}
		// Not whitespace, so must be the line count...
		m_uu_line_size = DEC(in);
		m_uu_state ++;
		break;
	case 1:	// We are awating the 1st char in the four char set
		m_uu_bytes[0] = in;
		m_uu_state ++;
		break;
	case 2: // We are awating the 2nd char in the four char set
		m_uu_bytes[1] = in;
		m_uu_state ++;
		break;
	case 3: // We are awating the 3rd char in the four char set
		m_uu_bytes[2] = in;
		m_uu_state ++;
		break;
	case 4: // We are awating the 4th char in the four char set
		m_uu_bytes[3] = in;
		int c[3];
		c[0] = DEC(m_uu_bytes[0]) << 2 | DEC(m_uu_bytes[1]) >> 4;
		c[1] = DEC(m_uu_bytes[1]) << 4 | DEC(m_uu_bytes[2]) >> 2;
		c[2] = DEC(m_uu_bytes[2]) << 6 | DEC(m_uu_bytes[3]);

		for (int i = 0; i < 3; i++)
		{
			if (m_uu_line_size > 0 && m_uu_size > 0)
			{
				*m_uu_data = c[i];
				m_uu_line_size --;
				m_uu_size --;
				m_uu_data ++;
			}
		}
		if (m_uu_line_size == 0)
		{
			m_uu_state = 0;
		}
		else
		{
			m_uu_state = 1;
		}
		break;
	}


}






// Type conversions....
void CXMLReader::unmakeString( CString str, CString &data )
{
	data = str;
}

void CXMLReader::unmakeString( CString str, int &data )
{
	data = atoi( str );
}

void CXMLReader::unmakeString( CString str, UINT &data )
{
	sscanf(str, "%u", &data );
}


void CXMLReader::unmakeString( CString str, COLORREF &data )
{
	char *dummy;
	data = strtol( str, &dummy, 16 );
}

void CXMLReader::unmakeString( CString str, double &data )
{
	data = atof( str );
}

void CXMLReader::unmakeString( CString str, CDPoint &data )
{
	sscanf( str, "%lg,%lg", &data.x, &data.y );
}

void CXMLReader::unmakeString( CString str, CPoint &data )
{
	CDPoint d;
	unmakeString( str, d );
	data.x = d.unmakeXMLUnits( d.x );
	data.y = d.unmakeXMLUnits( d.y );
}

void CXMLReader::unmakeString( CString str, BYTE &data )
{
	int d;
	unmakeString( str, d );
	data = d;
}

void CXMLReader::unmakeString( CString str, unsigned short &data )
{
	int d;
	unmakeString( str, d );
	data = d;
}

void CXMLReader::unmakeString( CString str, short &data )
{
	int d;
	unmakeString( str, d );
	data = d;
}

void CXMLReader::unmakeString( CString str, long &data )
{
	int d;
	unmakeString( str, d );
	data = d;
}
